# -*- coding: utf-8 -*-
"""Magnetic data collection.

This module collect the magnetic data from <combine_imps_wc> node, when the
the <combine_imps_w> combines two single impurity wc <kkr_imp_wc> or one
<combine_imps_wc> and <one kkr_imp_wc>.

Module includes two classes:
    CoupleImpurityData: The class collects the data of a couple of impurities
        from a specific <combine_imps_wc> node .

    MultiImpuritiesData: The class collects the data from all possible couples
    of the <combine_imps_wc> by implementing CoupleImpurityData class.
    And it  repeats the same process for all the <combine_imps_wc> nodes given
    by python list.
"""

from aiida.orm import Node
from typing import List
import math as m


class CoupleImpurityData(object):
    """A magnetic data of impurity couple.

    The class must be instantiated by combine_imps_wc 'node' and 'imp_orders'
    dict. Where the 'imp_orders' dict includes the values like 0,1 or 1,2 etc
    under any key name e.g. key1, key2. The numbers indicate the a couple
    first-second (0,1) impurities, second-third (1,2) impurities respectively.

    According to the magnetic impurity order the class collects data related
    to the impurities. The data includes J_ij interaction, magnetic moment for
    individual impurity, abs(DMI), DMI components, impurity distance
    components and so on.

    """

    def __init__(
        self,
        node: Node,
        imp_orders: dict,
    ) -> None:
        """Instantiate
        node: kkr_combine_imp node.
        imp_orders: Order of impurities to extract data.
                For example, for two impurities from 4-impurity
                calculation, the dict {key1: 1, key2: 2} indicates second
                and third impurities in the impurity list.
        """

        from aiida_kkr.workflows import combine_imps_wc

        self.J_data_uuid: str = None
        self.calc_imps: int = None
        self.imp1: int = None
        self.imp2: int = None
        self.offset1: int = None
        self.offset2: int = None
        self.il1: int = None
        self.il2: int = None
        self.J_ij: float = None
        self.Dx: float = None
        self.Dy: float = None
        self.Dz: float = None
        self.D: float = None
        self.rx: float = None
        self.ry: float = None
        self.rz: float = None
        self.r: float = None
        self.mom1: float = None
        self.mom2: float = None
        self.tot_mom: float = None

        # orders in imp pairs
        self.imp_order1: int = list(imp_orders.values())[0]
        self.imp_order2: int = list(imp_orders.values())[1]

        # To set the value for all attributes
        if issubclass(node.process_class, combine_imps_wc):
            self.ExtractData(node)
        else:
            raise TypeError(f'The node {node.uuid} is not a'
                            f'combine_imps_wc subclass')

        self.ExtractData(node)

    def __str__(self):
        print_str = (
            f'JijData_uuid: {self.J_data_uuid},\n'
            f'Total imps as part of calc: {self.calc_imps},\n'
            f'Zimp1: {self.imp1}, Zimp2: {self.imp2},\n'
            f'offset1: {self.offset1}, offset2: {self.offset2},\n'
            f'ilayer1: {self.il1}, ilayer2: {self.il2},\n'
            f'J_ij: {self.J_ij},\n'
            f'D_ij: {self.D}, Dx: {self.Dx}, Dy: {self.Dy},\n'
            f' Dz:{self.Dz},\n'
            f'r_ij: {self.r}, rx: {self.rx}, ry: {self.ry},\n'
            f'rz: {self.rz},\n'
            f'magnetic moment-1: {self.mom1},\n'
            f' magnetic moment-2: {self.mom2},\n'
            f'total moment: {self.tot_mom}'
        )
        return print_str

    def ExtractData(self, node: Node):
        """Extract data of magnetic impurity couple.

        node: Combine_imps_wc node and data will be extract from this node.
        """

        from aiida_kkr.workflows import kkr_imp_sub_wc

        imp_order1 = self.imp_order1
        imp_order2 = self.imp_order2
        workflow_info = node.outputs.workflow_info.get_dict()

        try:
            imps_info_in_exact_cluster = (workflow_info['imps_info_in_exact_cluster'])
        except KeyError:
            imps_info_in_exact_cluster = ExtractImpInfo(node)

        try:
            jij_data = node.outputs.JijData.get_array('JijData')
        except KeyError:
            raise KeyError(f'The combine_impurity_calc uuid: {node.uuid}'
                           f' did execute the Jij calculation.')

        else:
            jij_info = node.outputs.JijInfo.get_dict()['text']

        impurity_info = (node.get_outgoing(node_class=kkr_imp_sub_wc).all()[0].node.inputs.impurity_info.get_dict())
        zimp = impurity_info['Zimp']
        imp_cls = impurity_info['imp_cls']

        # To collect the impurity index from cluster
        zimp_index_cluster = [ind for ind, cluster_atom_info in enumerate(imp_cls) if cluster_atom_info[-2] in zimp]
        self.J_data_uuid = node.outputs.JijData.uuid
        self.calc_imps = len(zimp_index_cluster)
        self.imp1 = int(imp_cls[zimp_index_cluster[imp_order1]][-2])
        self.imp2 = int(imp_cls[zimp_index_cluster[imp_order2]][-2])

        # For old version of combine_imps_wc
        if len(zimp) != len(imps_info_in_exact_cluster['offset_imps']):
            self.offset1 = 0
            self.offset2 = imps_info_in_exact_cluster['offset_imps'][imp_order2 - 1]
        else:
            self.offset1 = imps_info_in_exact_cluster['offset_imps'][imp_order1]
            self.offset2 = imps_info_in_exact_cluster['offset_imps'][imp_order2]

        self.il1 = imps_info_in_exact_cluster['ilayers'][imp_order1]
        self.il2 = imps_info_in_exact_cluster['ilayers'][imp_order2]

        jij_info_arr = ([[float(data.strip())
                          for data in dataline.split(' ')
                          if data != '']
                         for dataline in jij_info.split('\n')[3:-1]])

        jij_pairs = [[data[0], data[1]] for data in jij_info_arr]
        jij_pair = [float(zimp_index_cluster[imp_order1]), float(zimp_index_cluster[imp_order2])]
        jij_pair_index = jij_pairs.index(jij_pair)
        r_j_d = jij_data[jij_pair_index][:]

        self.J_ij = r_j_d[3]

        self.D = r_j_d[4]
        self.Dx = r_j_d[5]
        self.Dy = r_j_d[6]
        self.Dz = r_j_d[7]

        self.rx = r_j_d[0]
        self.ry = r_j_d[1]
        self.rz = r_j_d[2]
        self.r = m.sqrt(sum([i**2 for i in r_j_d[0:3]]))

        # To collect the output data
        spin_moment_per_atom = (
            node.outputs.last_calc_output_parameters.get_dict()['magnetism_group']['spin_moment_per_atom']
        )
        self.mom1 = spin_moment_per_atom[int(jij_pair[0])][2]
        self.mom2 = spin_moment_per_atom[int(jij_pair[1])][2]
        self.tot_mom = self.mom1 + self.mom2


class MultiImpuritiesData(object):
    """Multiple impurity data extraction.

    This class extract magnetic data for all possible couples of impurities
    using CoupleImpurityData class. Later the process continue for all given
    nodes in node list. Then the data for all possible couples will be
    appended into dictionary as key->list map formate.
    """

    def __init__(self, node: Node):
        """Instantiate MultiImpurityData.

        inputs:
        node: <combine_imps_wc> node.
        """

        workflow_info = node.outputs.workflow_info.get_dict()
        self.zimps = workflow_info['imp_info_combined']['Zimp']

        self.node = node

        self.DataDictConstruct()
        self.DataDictFill()

    def DataDictConstruct(self):
        """Construct Data Dict.

        It construct the data dict in formate {property_name:list} that
        will be filled by the couple impurity data in DataDictFill method.
        """

        imp_num = len(self.zimps)
        multi_impurity_dict = dict()

        multi_impurity_dict['calc_imps'] = []
        multi_impurity_dict['J_data_uuid'] = []
        for imp_no in range(imp_num):
            multi_impurity_dict['imp' + str(imp_no)] = []
            multi_impurity_dict['offset' + str(imp_no)] = []
            multi_impurity_dict['ilayer' + str(imp_no)] = []

        multi_impurity_dict['i'] = []
        multi_impurity_dict['j'] = []
        multi_impurity_dict['Z_i'] = []
        multi_impurity_dict['Z_j'] = []
        multi_impurity_dict['J'] = []
        multi_impurity_dict['D'] = []
        multi_impurity_dict['Dx'] = []
        multi_impurity_dict['Dy'] = []
        multi_impurity_dict['Dz'] = []
        multi_impurity_dict['rx'] = []
        multi_impurity_dict['ry'] = []
        multi_impurity_dict['rz'] = []
        multi_impurity_dict['r'] = []
        multi_impurity_dict['mom1'] = []
        multi_impurity_dict['mom2'] = []
        multi_impurity_dict['tot_mom'] = []
        self.multi_impurity_dict = multi_impurity_dict

    def DataDictFill(self):
        """Fill Data Dict.

        Fill the data dict for all possible impurity couples.
        """

        multi_impurity_dict = self.multi_impurity_dict
        couple_imp_data_obj_list = []

        couple_imp_data_obj = None
        last_imp_order = None
        # Fill all the impurity filled but not interaction
        for imp1_order in range(len(self.zimps))[:-1]:
            couple_imp_data_obj = None
            for imp2_order in range(len(self.zimps))[imp1_order + 1:]:
                couple_imp_data_obj = CoupleImpurityData(
                    self.node, imp_orders={
                        'imp_order1': imp1_order,
                        'imp_order2': imp2_order
                    }
                )
                couple_imp_data_obj_list.append(couple_imp_data_obj)

                last_imp_order = imp2_order

            multi_impurity_dict['imp' + str(imp1_order)].append(couple_imp_data_obj.imp1)
            multi_impurity_dict['ilayer' + str(imp1_order)].append(couple_imp_data_obj.il1)
            multi_impurity_dict['offset' + str(imp1_order)].append(couple_imp_data_obj.offset1)

        multi_impurity_dict['imp' + str(last_imp_order)].append(couple_imp_data_obj.imp2)
        multi_impurity_dict['ilayer' + str(last_imp_order)].append(couple_imp_data_obj.il2)
        multi_impurity_dict['offset' + str(last_imp_order)].append(couple_imp_data_obj.offset2)

        # Here to fill up the data
        for enum, obj in enumerate(couple_imp_data_obj_list):
            multi_impurity_dict['J_data_uuid'].append(obj.J_data_uuid)
            multi_impurity_dict['calc_imps'].append(obj.calc_imps)
            multi_impurity_dict['i'].append(obj.imp_order1)
            multi_impurity_dict['j'].append(obj.imp_order2)
            multi_impurity_dict['Z_i'].append(obj.imp1)
            multi_impurity_dict['Z_j'].append(obj.imp2)
            multi_impurity_dict['J'].append(obj.J_ij)
            multi_impurity_dict['D'].append(obj.D)
            multi_impurity_dict['Dx'].append(obj.Dx)
            multi_impurity_dict['Dy'].append(obj.Dy)
            multi_impurity_dict['Dz'].append(obj.Dz)
            multi_impurity_dict['rx'].append(obj.rx)
            multi_impurity_dict['ry'].append(obj.ry)
            multi_impurity_dict['rz'].append(obj.rz)
            multi_impurity_dict['r'].append(obj.r)
            multi_impurity_dict['mom1'].append(obj.mom1)
            multi_impurity_dict['mom2'].append(obj.mom2)
            multi_impurity_dict['tot_mom'].append(obj.tot_mom)

            # To extend the impurity coloumn by one row not interaction
            if enum != len(couple_imp_data_obj_list) - 1:
                for imp_order in range(len(self.zimps)):

                    imp = multi_impurity_dict['imp' + str(imp_order)][-1]
                    il = multi_impurity_dict['ilayer' + str(imp_order)][-1]
                    offset = multi_impurity_dict['offset' + str(imp_order)][-1]

                    multi_impurity_dict['imp' + str(imp_order)].append(imp)
                    multi_impurity_dict['ilayer' + str(imp_order)].append(il)
                    multi_impurity_dict['offset' + str(imp_order)].append(offset)

        self.multi_impurity_dict = multi_impurity_dict

    def AppendDataMultipleNode(self, node_list: List[Node]):
        """Fill the Data Dict for a group of nodes.
       inputs:
        node_list: List[Node]
        Append the data to the Data Dict from a group of nodes given in
        'node_list' where each node created from the <combine_imps_wc> wf
        for same number of impurities.
        """

        if isinstance(node_list, list):
            for node in node_list:
                self.node = node
                self.DataDictFill()
        else:
            raise TypeError('Parameter node_list should be list of node.')

    @classmethod
    def ExtractDictMultipleNode(cls, node_list: List[Node]):
        """Use the MultiImpuritiesData class to extract data

        This classmethod will instantiate the class and collects
        magnetic data using instance method AppendDataMultipleNode(self).
        """

        from aiida_kkr.workflows import combine_imps_wc, kkr_imp_wc
        if issubclass(node_list[0].process_class, combine_imps_wc):
            obj = cls(node=node_list[0])
        else:
            raise TypeError(f'The node {node_list[0].uuid} is not a'
                            f'combine_imps_wc subclass')

        for node in node_list[1:]:
            if not issubclass(node_list[0].process_class, combine_imps_wc):
                raise TypeError(f'The node {node} in the node list not'
                                f' a {combine_imps_wc} class.')
        obj.AppendDataMultipleNode(node_list=node_list[1:])

        return obj

    def GetDataDict(self):
        """Return DataDict

        Return the Data Dict that is already filled.
        """
        return self.multi_impurity_dict

    def __repr__(self):
        """Print Data Dict.

        It helps for deburging step.
        """
        import pandas as pd
        multi_impurity_dict = self.multi_impurity_dict
        print_str = (f'multi impurity dict: \n{multi_impurity_dict}')
        return print_str


def ExtractImpInfo(node):
    """Extract impurity info from single kkrimp calc.

        This constructs and returns a python dict keeping info about two
        single impurities form <kkr_imp_wc> with respect to the original
        host structure e.i.
        before transforming the center to the first impurity position.

        Note: This ExtractImpInfo is required for the old combined_imps_wc
              for the calculation with the new version it irrelevant.
    """

    input_node_1 = node.inputs.impurity1_output_node
    input_node_2 = node.inputs.impurity2_output_node
    offset_imp2 = node.inputs.offset_imp2

    single_imp1_wc = get_imp_node_from_input(input_node_1)
    single_imp2_wc = get_imp_node_from_input(input_node_2)

    impinfo1 = single_imp1_wc.inputs.impurity_info
    impinfo2 = single_imp2_wc.inputs.impurity_info

    imps_info_in_exact_cluster = ({'Zimps': [], 'ilayers': [], 'offset_imps': [0]})

    #offset_imp contains offset_index for imps 2nd, 3rd so on
    zimp_1 = impinfo1.get_dict().get('Zimp')
    zimp_2 = impinfo2.get_dict().get('Zimp')

    if isinstance(zimp_1, list):
        zimp_1 = zimp_1[0]
    if isinstance(zimp_2, list):
        zimp_2 = zimp_2[0]

    imps_info_in_exact_cluster['Zimps'].append(zimp_1)
    imps_info_in_exact_cluster['Zimps'].append(zimp_2)
    imps_info_in_exact_cluster['ilayers'].append(impinfo1.get_dict().get('ilayer_center'))
    imps_info_in_exact_cluster['ilayers'].append(impinfo2.get_dict().get('ilayer_center'))
    imps_info_in_exact_cluster['offset_imps'].append(offset_imp2.get_dict().get('index'))

    return imps_info_in_exact_cluster


def get_imp_node_from_input(impurity_output_node):
    """ Extract parent node or calc node.
    This extracts the parent workflow node or KkrimpCalculation
    from 'impurity_output_node' given as input in combine_imps_wc
    workflow.
    """

    from aiida_kkr.calculations.kkrimp import KkrimpCalculation
    imp_out = impurity_output_node

    kkrimpcalc_parents = imp_out.get_incoming(node_class=KkrimpCalculation).all()
    if len(kkrimpcalc_parents) > 0:
        parent_imp1_wc_or_calc = kkrimpcalc_parents[0].node
    else:
        inc = imp_out.get_incoming(link_label_filter='workflow_info').all()
        if len(inc) != 1:
            raise ValueError('More input WorkChainNodes are found.')
        parent_imp1_wc_or_calc = inc[0].node

    return parent_imp1_wc_or_calc



# -*- coding: utf-8 -*-
"""
Here we specify some defaults for cli commands
"""


# Structures
def get_si_bulk_structure():
    """Return a `StructureData` representing bulk silicon.

    The database will first be queried for the existence of a bulk silicon crystal. If this is not the case, one is
    created and stored. This function should be used as a default for CLI options that require a `StructureData` node.
    This way new users can launch the command without having to construct or import a structure first. This is the
    reason that we hardcode a bulk silicon crystal to be returned. More flexibility is not required for this purpose.

    :return: a `StructureData` representing bulk silicon
    """
    from ase.spacegroup import crystal
    from aiida.orm import QueryBuilder, StructureData

    # Filters that will match any elemental Silicon structure with 2 or less sites in total
    filters = {
        'attributes.sites': {
            'of_length': 2
        },
        'attributes.kinds': {
            'of_length': 1
        },
        'attributes.kinds.0.symbols.0': 'Si'
    }

    builder = QueryBuilder().append(StructureData, filters=filters)
    results = builder.first()  # pylint: disable=no-value-for-parameter

    if not results:
        alat = 5.43
        ase_structure = crystal(
            'Si',
            [(0, 0, 0)],
            spacegroup=227,
            cellpar=[alat, alat, alat, 90, 90, 90],
            primitive_cell=True,
        )
        structure = StructureData(ase=ase_structure)
        structure.store()
    else:
        structure = results[0]  # pylint: disable=unsubscriptable-object

    return structure.uuid


def get_cu_bulk_structure():
    """Return a `StructureData` representing bulk copper.

    The database will first be queried for the existence of a bulk Cu crystal. If this is not the case, one is
    created and stored. This function should be used as a default for CLI options that require a `StructureData` node.
    This way new users can launch the command without having to construct or import a structure first. This is the
    reason that we hardcode a bulk silicon crystal to be returned. More flexibility is not required for this purpose.

    :return: a `StructureData` representing bulk silicon
    """
    from ase.spacegroup import crystal
    from aiida.orm import QueryBuilder, StructureData

    # Filters that will match any elemental Copper structure with 1 sites in total
    filters = {
        'attributes.sites': {
            'of_length': 1
        },
        'attributes.kinds': {
            'of_length': 1
        },
        'attributes.kinds.0.symbols.0': 'Cu'
    }

    builder = QueryBuilder().append(StructureData, filters=filters)
    results = builder.first()  # pylint: disable=no-value-for-parameter

    if not results:
        alat = 3.615
        ase_structure = crystal(
            'Cu',
            [(0, 0, 0)],
            spacegroup=225,
            cellpar=[alat, alat, alat, 90, 90, 90],
            primitive_cell=True,
        )
        structure = StructureData(ase=ase_structure)
        structure.store()
    else:
        structure = results[0]  # pylint: disable=unsubscriptable-object

    return structure.uuid


# Parameters
def get_kkrpara_defaults():
    """
    Create or load kkrparams default values
    """
    from aiida.orm import QueryBuilder, Dict
    from aiida_kkr.tools import kkrparams

    # Filters that will match the Dict node with kkr_default_params
    filters = {'extras.aiida-kkr_cmdline_info': 'kkr_default_params'}

    # now query for this
    builder = QueryBuilder().append(Dict, filters=filters)
    results = builder.first()  # pylint: disable=no-value-for-parameter

    # create kkr params node or take from database if found
    if not results:
        defaults, version = kkrparams.get_KKRcalc_parameter_defaults()
        params = Dict(defaults)
        params.label = 'Default KKR parameters'
        params.description = f'Version of the KKR params: {version}'
        params.extras['aiida-kkr_cmdline_info'] = 'kkr_default_params'  # pylint: disable=no-member
        params.store()
    else:
        params = results[0]  # pylint: disable=unsubscriptable-object

    # return params node
    return params.uuid


# Codes
def get_voro():
    """Return a `Code` node of the latest added inpgen executable in the database."""
    return get_last_code('kkr.voro')


def get_kkr():
    """Return a `Code` node of the latest added inpgen executable in the database."""
    return get_last_code('kkr.kkr')


def get_kkrimp():
    """Return a `Code` node of the latest added inpgen executable in the database."""
    return get_last_code('kkr.kkrimp')


def get_last_code(entry_point_name):
    """Return a `Code` node of the latest code executable of the given entry_point_name in the database.

    The database will be queried for the existence of a inpgen node.
    If this is not exists and NotExistent error is raised.


    :param entry_point_name: string
    :return: the uuid of a inpgen `Code` node
    :raise: aiida.common.exceptions.NotExistent
    """
    from aiida.orm import QueryBuilder, Code
    from aiida.common.exceptions import NotExistent

    filters = {'attributes.input_plugin': {'==': entry_point_name}}

    builder = QueryBuilder().append(Code, filters=filters)
    builder.order_by({Code: {'ctime': 'desc'}})
    results = builder.first()  # pylint: disable=no-value-for-parameter

    if not results:
        raise NotExistent(f'ERROR: Could not find any Code in the database with entry point: {entry_point_name}!')
    else:
        code = results[0]  # pylint: disable=unsubscriptable-object

    return code.uuid



# -*- coding: utf-8 -*-
"""
Parser for the KKR Code.
The parser should never fail, but it should catch
all errors and warnings and show them to the user.
"""

from aiida import __version__ as aiida_core_version
from aiida.parsers.parser import Parser
from aiida.orm import Dict
from aiida_kkr.calculations.kkr import KkrCalculation
from aiida.common.exceptions import InputValidationError, NotExistent
from masci_tools.io.parsers.kkrparser_functions import parse_kkr_outputfile, check_error_category
from masci_tools.io.common_functions import search_string
from aiida_kkr.tools.context import open_files_in_context
from contextlib import ExitStack

__copyright__ = (u'Copyright (c), 2017, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.8.0'
__contributors__ = ('Jens Broeder', u'Philipp Rüßmann')


class KkrParser(Parser):
    """
    Parser class for parsing output of KKR code..
    """

    def __init__(self, calc):
        """
        Initialize the instance of KkrParser
        """

        # needed for KKRimporter parser
        self.icrit = 0

        self._ParserVersion = __version__

        # reuse init of base class
        super(KkrParser, self).__init__(calc)

    # pylint: disable=protected-access

    def parse(self, debug=False, **kwargs):
        """
        Parse output data folder, store results in database.

        :param retrieved: a dictionary of retrieved nodes, where
          the key is the link name
        :returns: a tuple with two values ``(bool, node_list)``,
          where:

          * ``bool``: variable to tell if the parsing succeeded
          * ``node_list``: list of new nodes to be stored in the db
            (as a list of tuples ``(link_name, node)``)
        """
        success = False
        node_list = ()

        # Get retrieved folders
        try:
            out_folder = self.retrieved
        except NotExistent:
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        # check what is inside the folder
        list_of_files = out_folder.list_object_names()

        # we need at least the output file name as defined in calcs.py
        if KkrCalculation._DEFAULT_OUTPUT_FILE not in list_of_files:
            msg = f"Output file '{KkrCalculation._DEFAULT_OUTPUT_FILE}' not found in list of files: {list_of_files}"
            if self.icrit == 0:  # this check turns this off for the KKRimporter calculation
                self.logger.error(msg)
                return self.exit_codes.ERROR_NO_OUTPUT_FILE

        # determine whether or not everything should be parsed or not (e.g. qdos option)
        skip_mode = False
        only_000_present = False
        with out_folder.open(KkrCalculation._INPUT_FILE_NAME) as file:
            txt = file.readlines()
            itmp = search_string('RUNOPT', txt)
            if itmp >= 0:
                runopts = txt[itmp + 1].lower()
                if 'qdos' in runopts:
                    skip_mode = True
                if 'kkrflex' in runopts:
                    only_000_present = True

        # now collect the rest of the files
        file_errors = []

        # Parse output files of KKR calculation
        outfile_name = KkrCalculation._DEFAULT_OUTPUT_FILE
        if outfile_name not in out_folder.list_object_names():
            file_errors.append((1 + self.icrit, msg))
            outfile_name = None

        # get path to files and catch errors if files are not present
        # append tupels (error_category, error_message) where error_category is
        # 1: critical error, always leads to failing of calculation
        # 2: warning, is inspected and checked for consistency with read-in
        #    out_dict values (e.g. nspin, newsosol, ...)
        outfile_0init_name = KkrCalculation._OUTPUT_0_INIT
        if outfile_0init_name not in out_folder.list_object_names():
            file_errors.append((1 + self.icrit, f'Critical error! OUTPUT_0_INIT not found {outfile_0init_name}'))
            outfile_0init_name = None
        outfile_000_name = KkrCalculation._OUTPUT_000
        if outfile_000_name not in out_folder.list_object_names():
            file_errors.append((1 + self.icrit, f'Critical error! OUTPUT_000 not found {outfile_000_name}'))
            outfile_000_name = None
        outfile_2_name = KkrCalculation._OUTPUT_2
        if outfile_2_name not in out_folder.list_object_names():
            if not only_000_present:
                file_errors.append((1 + self.icrit, f'Critical error! OUTPUT_2 not found {outfile_2_name}'))
                outfile_2_name = None
            else:
                outfile_2_name = outfile_000_name
        potfile_out_name = KkrCalculation._OUT_POTENTIAL
        if potfile_out_name not in out_folder.list_object_names():
            file_errors.append((1 + self.icrit, f'Critical error! OUT_POTENTIAL not found {potfile_out_name}'))
            potfile_out_name = None
        timing_file_name = KkrCalculation._OUT_TIMING_000
        if timing_file_name not in out_folder.list_object_names():
            file_errors.append((1 + self.icrit, f'Critical error! OUT_TIMING_000  not found {timing_file_name}'))
            timing_file_name = None
        nonco_out_file_name = KkrCalculation._NONCO_ANGLES_OUT
        if nonco_out_file_name not in out_folder.list_object_names():
            file_errors.append((2, f'Error! NONCO_ANGLES_OUT not found {nonco_out_file_name}'))
            nonco_out_file_name = None

        out_dict = {
            'parser_version': self._ParserVersion,
            'calculation_plugin_version': KkrCalculation._CALCULATION_PLUGIN_VERSION
        }

        # TODO job title, compound description

        # open all files in context managers
        with ExitStack() as stack:
            # open files only if they exist
            # openend files are added to the context manager stack
            # so that at the end of the parsing all files are closed
            # when the context manager is exited
            (outfile, outfile_0init, outfile_000, outfile_2, potfile_out, timing_file,
             nonco_out_file) = open_files_in_context(
                 stack, out_folder, outfile_name, outfile_0init_name, outfile_000_name, outfile_2_name,
                 potfile_out_name, timing_file_name, nonco_out_file_name
             )

            # then parse the output
            out_dict = {}
            success, msg_list, out_dict = parse_kkr_outputfile(
                out_dict,
                outfile,
                outfile_0init,
                outfile_000,
                timing_file,
                potfile_out,
                nonco_out_file,
                outfile_2,
                skip_readin=skip_mode,
                debug=debug
            )

        # try to parse with other combinations of files to minimize parser errors
        if self.icrit != 0:
            self.logger.info(f'msg_list0: {msg_list}')
            # try second combination of files
            out_dict2 = out_dict.copy()
            success2, msg_list2, out_dict2 = parse_kkr_outputfile(
                out_dict2,
                outfile_2,
                outfile_0init,
                outfile_000,
                timing_file,
                potfile_out,
                nonco_out_file,
                outfile_2,
                skip_readin=skip_mode
            )
            self.logger.info(f'msg_list1: {msg_list2}')
            if len(msg_list2) < len(msg_list):  # overwrite parser outputs if fewer errors
                self.logger.info('take output of parser run 1')
                success, msg_list, out_dict = success2, msg_list2, out_dict2
            # try third combination of files
            out_dict2 = out_dict.copy()
            success2, msg_list2, out_dict2 = parse_kkr_outputfile(
                out_dict2,
                outfile_000,
                outfile_0init,
                outfile_000,
                timing_file,
                potfile_out,
                nonco_out_file,
                outfile_2,
                skip_readin=skip_mode
            )
            self.logger.info(f'msg_list2: {msg_list2}')
            if len(msg_list2) < len(msg_list):  # overwrite parser outputs if fewer errors
                self.logger.info('take output of parser run 2')
                success, msg_list, out_dict = success2, msg_list2, out_dict2

        # TODO discriminate between real errors and warnings
        msg_list = [i for i in msg_list if 'single particle energies' not in i]
        if len(msg_list) == 0:
            success = True

        out_dict['parser_errors'] = msg_list
        # add file open errors to parser output of error messages
        for (err_cat, f_err) in file_errors:
            if err_cat == 1:
                msg_list.append(f_err)
            elif check_error_category(err_cat, f_err, out_dict):
                msg_list.append(f_err)
            else:
                if 'parser_warnings' not in list(out_dict.keys()):
                    out_dict['parser_warnings'] = []
                out_dict['parser_warnings'].append(f_err.replace('Error', 'Warning'))
        out_dict['parser_errors'] = msg_list

        # create output node and link
        self.out('output_parameters', Dict(dict=out_dict))

        if self.icrit != 0 and not success:  # overwrite behavior with KKRimporter
            success = True  # set automatically to True even if only partial output was parsed
            msg = 'Automatically returned success=True for KKR importer although some parsing errors occurred'
            self.logger.warning(msg)

        if not success:
            return self.exit_codes.ERROR_KKR_PARSING_FAILED



# -*- coding: utf-8 -*-
"""
Helper functions used in KKR and voronoi parser. Needed to properly open file with the context manager.
"""


def open_context_to_stack(stack, out_folder, outfile_name, rw_mode=u'r'):
    """Open a file in a context manager which are collected
    in a context manager stack. The file is only opened if
    the filename is not None.

    :param stack:  a contextlib.ExitStack to which the opened file context is added
    :param out_folder: a retrieved folder for the calculation where the files are found
    :param outfile_name: name of the file that is supposed to be opened from the out_folder

    :returns: the file handle or None if the filename was None
    """
    if outfile_name is not None:
        outfile = stack.enter_context(out_folder.open(outfile_name, rw_mode))
    else:
        outfile = None
    return outfile


def open_files_in_context(stack, out_folder, *filenames):
    """Open files and add to context manager stack.

    :param stack:  a contextlib.ExitStack to which the opened file context is added
    :param out_folder: a retrieved folder for the calculation where the files are found
    :param filnames: list of filenames (should be strings) that are openend in out_folder

    :returns: a list of file handles
    """
    fhandles = []
    for fname in filenames:
        fhandles.append(open_context_to_stack(stack, out_folder, fname))

    return fhandles



#!/usr/bin/env python
# coding: utf-8
"""
This module contains the workflow that can be used to calculate the exchange coupling constants
"""

import numpy as np
from aiida.engine import WorkChain, ToContext, calcfunction
from aiida.orm import Dict, RemoteData, StructureData, ArrayData, CalcJobNode, Code
from aiida_kkr.tools.find_parent import get_calc_from_remote, find_parent_structure
from aiida_kkr.tools.jij_tools import parse_jij_calc, get_sites
from aiida_kkr.tools.common_workfunctions import test_and_get_codenode, get_parent_paranode, update_params_wf, get_inputs_kkr
from aiida_kkr.calculations.kkr import KkrCalculation
from aiida_kkr.tools.save_output_nodes import create_out_dict_node
from masci_tools.io.common_functions import get_alat_from_bravais
from masci_tools.io.kkr_params import kkrparams
from masci_tools.util.constants import BOHR_A

__copyright__ = (u'Copyright (c), 2022, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.1.2'
__contributors__ = (u'Philipp Rüßmann')


class kkr_jij_wc(WorkChain):
    """
    Workchain for calculation of exchange coupling constants Jij and Dij if parent calculation used the SOC solver.

    inputs::

        :param wf_parameters: optional Dict node of workchain specifications, contains settings like Jij radius cutoff,
                              selection of sites for i and j and numerical cutoffs. None values in the accuracy sub-dict
                              means that values from parent calculation are coptied.
        :param remote_data: mandatory RemoteData node of parent (i.e. converged) KkrCalculation
        :param kkr: optional Code for KKRhost executable (if not given the same as in the parent calculation is used)
        :param options: optional Dict computer options like scheduler command or parallelization

    returns::

        :return jij_data: ArrayData with the arrays 'Jij_expanded' (Table of all Jij and Dij pairs) and 'positions_expanded' (positions of all ij pairs)
        :return structure_jij_sites: StructureData
    """

    _wf_version = __version__
    _wf_default = {
        'jijrad_ang': 5.0, # default cutoff radius
        'jijsite_i': None, # use all sites by default
        'jijsite_j': None, # use all sites by default
        'accuracy': { # accuracy settings, typically set to larger values than in scf run
            'NATOMIMPD': 500,
            'NSHELD': 2000,
            'TEMPR': None,
            'RCLUSTZ': None,
            'kmesh': None,
        },
    }
    _options_default = {
        'max_wallclock_seconds': 36000,
        'resources': {
            'num_machines': 1
        },
        'withmpi': True,
        'queue_name': ''
    }

    @classmethod
    def get_wf_defaults(self, silent=False):
        """
        Return the default values of the workflow parameters (wf_parameters input node)
        """
        if not silent:
            print(f'Version of the kkr_jij_wc workflow: {self._wf_version}')
        return self._wf_default.copy()

    @classmethod
    def define(cls, spec):
        """
        Layout of the workflow, defines the input nodes and the outline of the workchain
        """
        super(kkr_jij_wc, cls).define(spec)

        # here inputs are defined
        spec.input(
            'wf_parameters',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._wf_default),
            help='Parameters of the bandstructure workflow (see output of kkr_bs_wc.get_wf_default() for more details).'
        )

        spec.input(
            'options',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._options_default),
            help=
            'Computer options (walltime etc.) passed onto KkrCalculation, fall back to settings from parent calculation if not given'
        )

        spec.input(
            'remote_data',
            valid_type=RemoteData,
            required=True,
            help='Parent folder of previously converged KkrCalculation'
        )

        spec.input('kkr', valid_type=Code, required=True, help='KKRhost code, needed to run the Jij KkrCalculation')

        spec.input(
            'params_kkr_overwrite',
            valid_type=Dict,
            required=False,
            help='Overwrite some input parameters of the parent KKR calculation.'
        )

        # Here outputs are defined
        spec.output('results_wf', valid_type=Dict, required=True)
        spec.output('jij_data', valid_type=ArrayData, required=True)
        spec.output('structure_jij_sites', valid_type=StructureData, required=True)

        # Here outlines are being specified
        spec.outline(
            # For initialiging workflow
            cls.start,
            cls.validate_input,
            cls.set_jij_params,
            cls.submit_Jij_calcs,
            cls.return_results
        )
        # definition of exit code in case something goes wrong in this workflow
        spec.exit_code(
            160, 'ERROR_KKRCODE_NOT_CORRECT', 'The code you provided for kkr does not use the plugin kkr.kkr'
        )
        spec.exit_code(161, 'ERROR_INVALID_PARENT', 'Parent calculation is not valid')
        spec.exit_code(162, 'ERROR_CALC_FAILED', 'KKR Band Structure calculation failed')
        spec.exit_code(163, 'ERROR_PARSING_FAILED', 'Parsing of Jij calculations failed')

    def start(self):
        """
        set up context of the workflow
        """
        self.report(f'INFO: started KKR Jij workflow version {self._wf_version}')
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = {}
        # add missing default valuesi
        for key, val in self._wf_default.items():
            if ((key not in wf_dict.keys()) and (key.swapcase() not in wf_dict.keys()) and (val is not None)):
                self.report(f'INFO: Using default wf parameter {key}: {val}')
                wf_dict[key] = val

        if 'options' in self.inputs:
            options_dict = self.inputs.options.get_dict()
        else:
            options_dict = self._options_default
        self.ctx.options = options_dict

        self.ctx.withmpi = options_dict.get('withmpi', self._options_default['withmpi'])
        self.ctx.resources = options_dict.get('resources', self._options_default['resources'])
        self.ctx.max_wallclock_seconds = options_dict.get(
            'max_wallclock_seconds', self._options_default['max_wallclock_seconds']
        )
        self.ctx.queue = options_dict.get('queue_name', self._options_default['queue_name'])
        self.ctx.custom_scheduler_commands = options_dict.get('custom_scheduler_commands', '')
        self.ctx.wf_dict = wf_dict
        self.report(
            'INFO: use the following parameter:\n'
            'withmpi: {}\n'
            'Resources: {}\n'
            'Walltime (s): {}\n'
            'queue name: {}\n'
            'scheduler command: {}\n'
            'Workflow parameters: {}\n'.format(
                self.ctx.withmpi, self.ctx.resources, self.ctx.max_wallclock_seconds, self.ctx.queue,
                self.ctx.custom_scheduler_commands, wf_dict
            )
        )

    def validate_input(self):
        """
        validate inputs
        """

        # save parent calculation
        input_remote = self.inputs.remote_data
        parents = input_remote.get_incoming(node_class=CalcJobNode).all()
        if len(parents) != 1:
            # check if parent is unique
            return self.exit_codes.ERROR_INVALID_PARENT  # pylint: disable=no-member
        self.ctx.parent_calc = get_calc_from_remote(input_remote)

        # validate for kkrcode
        try:
            test_and_get_codenode(self.inputs.kkr, 'kkr.kkr', use_exceptions=True)
        except ValueError:
            return self.exit_codes.ERROR_KKRCODE_NOT_CORRECT  # pylint: disable=no-member

        # get parent input parameter
        self.ctx.input_params_KKR = get_parent_paranode(self.inputs.remote_data)

    def set_jij_params(self):
        """
        set kkr parameters for the Jij calculation
        """

        # input parameters from parent
        params = self.ctx.input_params_KKR

        # maybe overwrite some inputs
        if 'params_kkr_overwrite' in self.inputs:
            self.report(f'found params_kkr_overwrite: {self.inputs.params_kkr_overwrite.get_dict()}')
            updatenode = self.inputs.params_kkr_overwrite
            updatenode.label = 'params overwrite'
            params = update_params_wf(params, updatenode)

        # set Jij parameters
        para_jij, runopts = self._get_para_jij(params)
        updatenode = Dict(para_jij.get_dict())
        updatenode.label = 'Jij params'
        paranode_jij = update_params_wf(params, updatenode)
        self.ctx.jij_params = paranode_jij

        # find out if we have a calculation with or without SOC (then no DMI is calculated)
        self.ctx.noSOC = True
        if 'NEWSOSOL' in runopts or para_jij.get_value('<USE_CHEBYCHEV_SOLVER>'):
            self.ctx.noSOC = False

    def submit_Jij_calcs(self):
        """
        submit the KkrCalcultion with the Jij settings
        """

        # get inputs for band structure calculation
        inputs = get_inputs_kkr(
            self.inputs.kkr,
            self.inputs.remote_data,
            self.ctx.options,
            label='Jij_calc',
            description='',
            parameters=self.ctx.jij_params,
            serial=(not self.ctx.withmpi)
        )

        # noSOC, only m||z
        if self.ctx.noSOC:
            inputs.metadata.label = 'jij_calc_z'  # pylint: disable=no-member
            jij_calc_z = self.submit(KkrCalculation, **inputs)
            self.ctx.jij_calc_z = jij_calc_z
            self.ctx.jij_calc_x = None
            self.ctx.jij_calc_y = None
        else:
            # create nonco angles for the three calculations
            nonco_angles = _make_nonco_angles(parent_remote=self.inputs.remote_data)
            init_angles_x, init_angles_y, init_angles_z = nonco_angles['init_angles_x'], nonco_angles[
                'init_angles_y'], nonco_angles['init_angles_z']

            # submit m||z calculation
            inputs.initial_noco_angles = init_angles_z
            inputs.metadata.label = 'jij_calc_z'  # pylint: disable=no-member
            jij_calc_z = self.submit(KkrCalculation, **inputs)
            self.ctx.jij_calc_z = jij_calc_z

            # submit m||x calculation
            inputs.initial_noco_angles = init_angles_x
            inputs.metadata.label = 'jij_calc_x'  # pylint: disable=no-member
            jij_calc_x = self.submit(KkrCalculation, **inputs)
            self.ctx.jij_calc_x = jij_calc_x

            # submit m||y calculation
            inputs.initial_noco_angles = init_angles_y
            inputs.metadata.label = 'jij_calc_y'  # pylint: disable=no-member
            jij_calc_y = self.submit(KkrCalculation, **inputs)
            self.ctx.jij_calc_y = jij_calc_y

        # add to context (needed to tell aiida to wait for processes to finish)
        futures = {'jij_calc_z': self.ctx.jij_calc_z}
        if self.ctx.jij_calc_x is not None:
            futures['jij_calc_x'] = self.ctx.jij_calc_x
        if self.ctx.jij_calc_y is not None:
            futures['jij_calc_y'] = self.ctx.jij_calc_y
        return ToContext(**futures)

    def return_results(self):
        """
        Collect results, parse Jij output and link output nodes to workflow node
        """

        # check if calculations finished ok
        success = True
        if self.ctx.jij_calc_z is not None and not self.ctx.jij_calc_z.is_finished_ok:
            success = False
        if self.ctx.jij_calc_x is not None and not self.ctx.jij_calc_x.is_finished_ok:
            success = False
        if self.ctx.jij_calc_y is not None and not self.ctx.jij_calc_y.is_finished_ok:
            success = False
        if not success:
            self.ctx.successful = False
            error = f'ERROR Jij calculation failed somehow it is in state {self.ctx.jij_calc_z.process_state}'
            if self.ctx.jij_calc_x is not None:
                error += f'; {self.ctx.jij_calc_x.process_state} (x)'
            if self.ctx.jij_calc_y is not None:
                error += f'; {self.ctx.jij_calc_y.process_state} (y)'
            self.report(error)
            return self.exit_codes.ERROR_CALC_FAILED  # pylint: disable=no-member

        # now parse calculation output
        try:
            jij_data, structure_jij_sites = parse_jij_calc(
                self.ctx.jij_calc_z, jij_calc_x=self.ctx.jij_calc_x, jij_calc_y=self.ctx.jij_calc_y, verbose=False
            )
        except Exception as err:
            self.report(f'Caught error when trying to parse Jij output:{err}')
            return self.exit_codes.ERROR_PARSING_FAILED  # pylint: disable=no-member

        # collect output nodes
        outdict = {'jij_data': jij_data, 'structure_jij_sites': structure_jij_sites}

        # create dict to store results of workflow output
        outputnode_dict = {}
        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._wf_version
        outputnode_dict['successful'] = success

        # create output node with data-provenance
        outputnode = Dict(outputnode_dict)

        # link to the output nodes
        link_nodes = outdict.copy()

        outdict['results_wf'] = create_out_dict_node(outputnode, **link_nodes)

        # create links to output nodes
        for link_name, node in outdict.items():
            self.out(link_name, node)

        self.report('INFO: done with Jij workflow!')

    # Helper functions

    def _get_para_jij(self, params):
        """
        Set the Jij parameters from the input.
        Returns a kkrparams instance with the set values
        """

        # get input parameters
        input_dict = params.get_dict()
        para_jij = kkrparams(**input_dict)

        # set Jij parameters
        # add 'XCPL' runopt to list of runopts (activates Jij calculation)
        runopts = input_dict.get('RUNOPT')
        if runopts is None:
            runopts = []
        runopts.append('XCPL    ')
        para_jij.set_value('RUNOPT', runopts)
        para_jij.set_value('NSTEPS', 1)  # one-shot run

        # accuracy settings
        tempr = self.ctx.wf_dict.get('accuracy', {}).get('TEMPR')
        if tempr is not None:
            para_jij.set_value('TEMPR', tempr)  # slightly reduce temperature

        rclustz = self.ctx.wf_dict.get('accuracy', {}).get('RCLUSTZ')
        if rclustz is not None:
            para_jij.set_value('RCLUSTZ', rclustz)  # increase cluster radius

        kmesh = self.ctx.wf_dict.get('accuracy', {}).get('kmesh')
        if kmesh is not None:
            para_jij.set_value('BZDIVIDE', kmesh)  # increase k-points
            para_jij.set_value('KPOIBZ', np.product(kmesh))  # array dimension

        # array dimensions
        NATOMIMPD = self.ctx.wf_dict.get('accuracy', {}).get('NATOMIMPD')
        if NATOMIMPD is not None:
            para_jij.set_value('NATOMIMPD', NATOMIMPD)  # array dimension

        NSHELD = self.ctx.wf_dict.get('accuracy', {}).get('NSHELD')
        if NSHELD is not None:
            para_jij.set_value('NSHELD', NSHELD)  # array dimension

        # Jij settings
        jijrad = self._get_jijrad()
        if jijrad is not None:
            para_jij.set_value('JIJRAD', jijrad)  # radius in lattice constants up to which the Jijs are calculated

        # set optional Jij parameters
        # i and j index for Jij calculation in internal units
        # uses site index (i.e. needs to be <=10)
        JIJSITEI = self.ctx.wf_dict.get('jijsite_i')
        JIJSITEJ = self.ctx.wf_dict.get('jijsite_j')
        if JIJSITEI is not None:
            if JIJSITEJ is None:
                JIJSITEJ = JIJSITEI
            para_jij.set_value('JIJSITEI', [len(JIJSITEI)] + [i + 1 for i in JIJSITEI])
        if JIJSITEJ is not None:
            para_jij.set_value('JIJSITEJ', [len(JIJSITEJ)] + [i + 1 for i in JIJSITEJ])

        return para_jij, runopts

    def _get_jijrad(self):
        """
        get Jij radius convert from Ang to internal alat units
        """

        # Jij radius in Ang.
        jijrad_ang = self.ctx.wf_dict.get('jijrad_ang')

        # find structure from calculation
        struc, _ = find_parent_structure(self.ctx.parent_calc)
        self.ctx.structure = struc

        # get alat from structure
        alat_ang = get_alat_from_bravais(np.array(struc.cell), struc.pbc[2])

        # maybe use value provided in input instead
        para = {k.lower(): v for k, v in self.ctx.parent_calc.inputs.parameters.get_dict().items() if v is not None}
        if para.get('use_alat_input', False) or para.get('use_input_alat', False):
            alat_ang = para.get('alatbasis') * BOHR_A

        # now have Jij radius in alat units
        jijrad = jijrad_ang / alat_ang

        return jijrad


@calcfunction
def _make_nonco_angles(parent_remote):
    """
    Create nonco angles for the 3 directions (x y, z)
    """
    # find structure to count number of sites
    structure = find_parent_structure(parent_remote)[0]
    Nsites = len(get_sites(structure))

    # create nonco angles for m||z
    init_angles_z = Dict({
        'fix_dir': [True for i in range(Nsites)],
        'theta': [0.0 for i in range(Nsites)],
        'phi': [0.0 for i in range(Nsites)],
    })

    # create nonco angles for m||x
    init_angles_x = Dict({
        'fix_dir': [True for i in range(Nsites)],
        'theta': [90.0 for i in range(Nsites)],
        'phi': [0.0 for i in range(Nsites)],
    })

    # create nonco angles for m||y
    init_angles_y = Dict({
        'fix_dir': [True for i in range(Nsites)],
        'theta': [90.0 for i in range(Nsites)],
        'phi': [90.0 for i in range(Nsites)],
    })

    return {'init_angles_x': init_angles_x, 'init_angles_y': init_angles_y, 'init_angles_z': init_angles_z}



# -*- coding: utf-8 -*-
"""
Options commonly used throughout the aiida-kkr command line interface.
To standardize at least of the options, they are kept as close as possible to
aiida-fleur. aiida-core and aiida-quantumespresso
"""
import click
from aiida.cmdline.params import types
from aiida.cmdline.params.options import OverridableOption
from .types import StructureNodeOrFileParamType
from .defaults import get_voro, get_kkr, get_kkrimp, get_kkrpara_defaults

STRUCTURE_OR_FILE = OverridableOption(
    '-s',
    '--structure',
    type=StructureNodeOrFileParamType(),
    help='StructureData node, given by pk or uuid or file in any for mat which will be converted.'
)

STRUCTURE = OverridableOption(
    '-s',
    '--structure',
    type=types.DataParamType(sub_classes=('aiida.data:core.structure',)),
    help='StructureData node, given by pk or uuid.'
)

VORO = OverridableOption(
    '-v',
    '--voro',
    type=types.CodeParamType(entry_point='kkr.voro'),
    default=get_voro,
    show_default=True,
    help='A code node or label for an voronoi executable.'
)

KKR = OverridableOption(
    '-k',
    '--kkr',
    type=types.CodeParamType(entry_point='kkr.kkr'),
    default=get_kkr,
    show_default=True,
    help='A code node or label for a KKR executable.'
)

KKRIMP = OverridableOption(
    '-ki',
    '--kkr-imp',
    type=types.CodeParamType(entry_point='kkr.kkrimp'),
    default=get_kkrimp,
    show_default=True,
    help='A code node or label for a KKRimp executable.'
)

PARAMETERS = OverridableOption(
    '-p',
    '--parameters',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    default=get_kkrpara_defaults,
    show_default=True,
    help='Dict with calculation parameters which will be given to the calculation or workchain.'
)

KPOINTS = OverridableOption(
    '-kpt',
    '--kpoints',
    type=types.DataParamType(sub_classes=('aiida.data:core.array.kpoints',)),
    help='An aiida KpointsData node.'
)

POTENTIAL_OVERWRITE = OverridableOption(
    '--potential-overwrite',
    type=types.DataParamType(sub_classes=('aiida.data:core.singlefile',)),
    help='Use a node that specifies the potential which is used instead of the voronoi output potential'
)

IMPURITY_INFO = OverridableOption(
    '-i',
    '--impurity-info',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    help=
    'Dict containing parameters that specify properties for a following impurity calculation (e.g. setting of impurity cluster in scoef file that is automatically created'
)

WF_PARAMETERS = OverridableOption(
    '-wf',
    '--wf-parameters',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    help='Dict containing parameters given to the workchain.'
)

VORO_PARAMETERS = OverridableOption(
    '-vp',
    '--voro-parameters',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    help='Dict containing parameters for the auxiliary voronoi starting potential workflow.'
)

OPTION_NODE = OverridableOption(
    '-opt',
    '--option-node',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    help=
    'Dict, an option node for the workchain containing for instance: {"withmpi": False, "max_wallclock_seconds": 6000, "resources": {"num_machines": 1, "num_mpiprocs_per_machine": 1}}'
)

MAX_NUM_MACHINES = OverridableOption(
    '-N',
    '--max-num-machines',
    type=click.INT,
    default=1,
    show_default=True,
    help='The maximum number of machines (nodes) to use for the calculations.'
)

MAX_WALLCLOCK_SECONDS = OverridableOption(
    '-W',
    '--max-wallclock-seconds',
    type=click.INT,
    default=1800,
    show_default=True,
    help='The maximum wallclock time in seconds to set for the calculations.'
)

NUM_MPIPROCS_PER_MACHINE = OverridableOption(
    '-M',
    '--num-mpiprocs-per-machine',
    type=click.INT,
    default=12,
    show_default=True,
    help='Run the simulation with so many num-mpi-procs-per-machine.'
)

WITH_MPI = OverridableOption(
    '-I', '--with-mpi', is_flag=True, default=False, show_default=True, help='Run the calculations with MPI enabled.'
)

QUEUE_NAME = OverridableOption(
    '-Q',
    '--queue-name',
    default='',
    show_default=True,
    help='The queue_name to be used in the submission script of the job.'
)

PARENT_FOLDER = OverridableOption(
    '-P',
    '--parent-folder',
    'parent_folder',
    type=types.DataParamType(sub_classes=('aiida.data:core.remote',)),
    show_default=True,
    required=False,
    help='The PK of a parent remote folder (for restarts).'
)

DAEMON = OverridableOption(
    '-d',
    '--daemon',
    is_flag=True,
    default=False,
    show_default=True,
    help='Submit the process to the daemon instead of running it locally. -d flag does not need any argument'
)

WF_PARAMETERS = OverridableOption(
    '-wf',
    '--wf-parameters',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    help='Dict containing parameters given to the workchain.'
)

VORO_PARAMETERS = OverridableOption(
    '-vp',
    '--voro-parameters',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    help='Dict containing parameters for the auxiliary voronoi starting potential workflow.'
)

OPTION_NODE = OverridableOption(
    '-opt',
    '--option-node',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    help='Dict, an option node for the workchain.'
)

MAX_NUM_MACHINES = OverridableOption(
    '-N',
    '--max-num-machines',
    type=click.INT,
    default=1,
    show_default=True,
    help='The maximum number of machines (nodes) to use for the calculations.'
)

MAX_WALLCLOCK_SECONDS = OverridableOption(
    '-W',
    '--max-wallclock-seconds',
    type=click.INT,
    default=1800,
    show_default=True,
    help='The maximum wallclock time in seconds to set for the calculations.'
)

NUM_MPIPROCS_PER_MACHINE = OverridableOption(
    '-M',
    '--num-mpiprocs-per-machine',
    type=click.INT,
    default=12,
    show_default=True,
    help='Run the simulation with so many num-mpi-procs-per-machine.'
)

WITH_MPI = OverridableOption(
    '-I', '--with-mpi', is_flag=True, default=False, show_default=True, help='Run the calculations with MPI enabled.'
)

PARENT_FOLDER = OverridableOption(
    '-P',
    '--parent-folder',
    'parent_folder',
    type=types.DataParamType(sub_classes=('aiida.data:core.remote',)),
    show_default=True,
    required=False,
    help='The PK of a parent remote folder (for restarts).'
)

DAEMON = OverridableOption(
    '-d',
    '--daemon',
    is_flag=True,
    default=False,
    show_default=True,
    help='Submit the process to the daemon instead of running it locally.'
)

NOCO_ANGLES = OverridableOption(
    '--noco-angles',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    help='Dict containing the non-collinear angles for the magnetic moments. See KkrCalculation for details.'
)



# -*- coding: utf-8 -*-
"""
Input plug-in for a KKR calculation.
"""
import os
import numpy as np
from aiida.engine import CalcJob
from aiida.orm import CalcJobNode, load_node, RemoteData, Dict, StructureData, KpointsData, Bool, FolderData
from .voro import VoronoiCalculation
from ..tools.common_workfunctions import get_natyp
from aiida.common.utils import classproperty
from aiida.common.exceptions import InputValidationError, ValidationError
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.common.exceptions import UniquenessError
from aiida_kkr.tools import (
    generate_inputcard_from_structure, check_2Dinput_consistency, vca_check, kick_out_corestates
)
from aiida_kkr.tools.find_parent import find_parent_structure
from aiida_kkr.tools.tools_kkrimp import make_scoef, write_scoef_full_imp_cls
from aiida_kkr.tools.ldau import get_ldaupot_text
from masci_tools.io.common_functions import get_alat_from_bravais, get_Ang2aBohr, search_string, get_ef_from_potfile
from masci_tools.io.kkr_params import __kkr_default_params__, kkrparams

__copyright__ = (u'Copyright (c), 2017, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'

__version__ = '0.13.2'

__contributors__ = ('Jens Bröder', 'Philipp Rüßmann')

verbose = False


class KkrCalculation(CalcJob):
    """
    AiiDA calculation plugin for a KKR calculation.
    """

    # calculation plugin version
    _CALCULATION_PLUGIN_VERSION = __version__

    # Default input and output files
    _DEFAULT_INPUT_FILE = 'inputcard'  # will be shown with inputcat
    _DEFAULT_OUTPUT_FILE = 'out_kkr'  # verdi shell output will be shown with outputcat

    # same as _DEFAULT_OUTPUT_FILE: piped output of kkr execution to this file
    _OUTPUT_FILE_NAME = _DEFAULT_OUTPUT_FILE

    # List of mandatory input files
    _INPUT_FILE_NAME = _DEFAULT_INPUT_FILE
    _POTENTIAL = 'potential'

    # List of optional input files (may be mandatory for some settings in inputcard)
    _SHAPEFUN = 'shapefun'  # mandatory if nonspherical calculation
    _SCOEF = 'scoef'  # mandatory for KKRFLEX calculation and some functionalities
    _BFIELD = 'bfield.dat'  # mandatory if <NONCOBFIELD>= True
    _NONCO_ANGLES = 'nonco_angle.dat'  # mandatory if noncollinear directions are used that are not (theta, phi)= (0,0) for all atoms
    _NONCO_ANGLES_IMP = 'nonco_angle_imp.dat'  # mandatory for GREENIMP option (scattering code)
    _SHAPEFUN_IMP = 'shapefun_imp'  # mandatory for GREENIMP option (scattering code)
    _POTENTIAL_IMP = 'potential_imp'  # mandatory for GREENIMP option (scattering code)

    # List of output files that should always be present
    _OUT_POTENTIAL = 'out_potential'
    _OUTPUT_0_INIT = 'output.0.txt'
    _OUTPUT_000 = 'output.000.txt'
    _OUTPUT_2 = 'output.2.txt'
    _OUT_TIMING_000 = 'out_timing.000.txt'
    _NONCO_ANGLES_OUT = 'nonco_angle_out.dat'
    _NONCO_ANGLES_ALL_ITER = 'nonco_angle_out_all_iter.dat'

    # special files (some runs)
    # DOS files
    _COMPLEXDOS = 'complex.dos'
    _DOS_ATOM = 'dos.atom%i'
    _LMDOS = 'lmdos.%2i.%i.dat'
    # qdos files
    _QVEC = 'qvec.dat'
    _QDOS_ATOM = 'qdos.%3i.%i.dat'
    _QDOS_ATOM_OLD = 'qdos.%2i.%i.dat'
    _QDOS_SX = 'qdos_sx.%2i.dat'
    _QDOS_SY = 'qdos_sy.%2i.dat'
    _QDOS_SZ = 'qdos_sz.%2i.dat'
    # kkrflex files for impurity calculation
    _KKRFLEX_GREEN = 'kkrflex_green'
    _KKRFLEX_TMAT = 'kkrflex_tmat'
    _KKRFLEX_ATOMINFO = 'kkrflex_atominfo'
    _KKRFLEX_INTERCELL_REF = 'kkrflex_intercell_ref'
    _KKRFLEX_INTERCELL_CMOMS = 'kkrflex_intercell_cmoms'
    _ALL_KKRFLEX_FILES = [
        _KKRFLEX_GREEN,
        _KKRFLEX_TMAT,
        _KKRFLEX_ATOMINFO,
        _KKRFLEX_INTERCELL_REF,
        _KKRFLEX_INTERCELL_CMOMS,
    ]
    # Jij files
    _Jij_ATOM = 'Jij.atom%0.5i'
    _SHELLS_DAT = 'shells.dat'
    # deci-out and decimation
    _DECIFILE = 'decifile'
    # BdG mode
    _BDG_POT = 'den_lm_ir.%0.3i.%i.txt'
    _BDG_CHI_NS = 'den_lm_ns_*.npy'
    # LDA+U
    _LDAUPOT = 'ldaupot'

    # template.product entry point defined in setup.json
    _default_parser = 'kkr.kkrparser'

    # list of keywords that are not allowed to be modified (new calculation
    # starting from structure and voronoi run is needed instead):
    _do_never_modify = [
        'ALATBASIS',
        'BRAVAIS',
        'NAEZ',
        '<RBASIS>',
        'CARTESIAN',
        'INTERFACE',
        '<NLBASIS>',
        '<RBLEFT>',
        'ZPERIODL',
        '<NRBASIS>',
        '<RBRIGHT>',
        'ZPERIODR',
        'KSHAPE',
        '<SHAPE>',
        '<ZATOM>',
        'NATYP',
        '<SITE>',
        '<CPA-CONC>',
        '<KAOEZL>',
        '<KAOEZR>',
    ]
    #TODO implement workfunction to modify structure (e.g. to use VCA)

    # small number used to check for equivalence
    _eps = 10**-12

    @classmethod
    def define(cls, spec):
        """
        Init internal parameters at class load time
        """
        # reuse base class function
        super(KkrCalculation, cls).define(spec)

        # now define input files and parser
        spec.input(
            'metadata.options.parser_name',
            valid_type=str,
            default=cls._default_parser,
            non_db=True,
        )
        spec.input('metadata.options.input_filename', valid_type=str, default=cls._DEFAULT_INPUT_FILE, non_db=True)
        spec.input('metadata.options.output_filename', valid_type=str, default=cls._DEFAULT_OUTPUT_FILE, non_db=True)

        # define input nodes (optional ones have required=False)
        spec.input(
            'parameters',
            valid_type=Dict,
            required=True,
            help='Use a node that specifies the input parameters',
        )
        spec.input(
            'parent_folder',
            valid_type=RemoteData,
            required=True,
            help="""
Use a remote or local repository folder as parent folder
(also for restarts and similar). It should contain all the  needed
files for a KKR calc, only edited files should be uploaded from the
repository.
"""
        )
        spec.input(
            'impurity_info',
            valid_type=Dict,
            required=False,
            help="""
Use a Parameter node that specifies properties for a following
impurity calculation (e.g. setting of impurity cluster in scoef
file that is automatically created)."""
        )
        spec.input(
            'kpoints',
            valid_type=KpointsData,
            required=False,
            help="""
Use a KpointsData node that specifies the kpoints for which a
bandstructure (i.e. 'qdos') calculation should be performed."""
        )
        spec.input(
            'initial_noco_angles',
            valid_type=Dict,
            required=False,
            help="""
Initial non-collinear angles for the magnetic moments of
the impurities. These values will be written into the
`kkrflex_angle` input file of KKRimp.
The Dict node should be of the form
initial_noco_angles = Dict(dict={
    'theta': [theta_at1, theta_at2, ..., theta_atN],
    # list theta values in degrees (0..180)
    'phi': [phi_at1, phi_at2, ..., phi_atN],
    # list phi values in degrees (0..360)
    'fix_dir': [True/False at_1, ..., True/False at_N]
    # list of booleans indicating if the direction of the magnetic
    # moment should be fixed or is allowed relax (True means keep the
    # direction of the magnetic moment fixed)
})
Note: The length of the theta, phi and fix_dir lists have to be
equal to the number of atoms.
"""
        )
        spec.input(
            'bfield',
            valid_type=Dict,
            required=False,
            help="""
Non-collinear exteral B-field used for constraint calculations.

The Dict node should be of the form
initial_noco_angles = Dict(dict={
    'theta': [theta_at1, theta_at2, ..., theta_atN],
    # list theta values in degrees (0..180)
    'phi': [phi_at1, phi_at2, ..., phi_atN],
    # list phi values in degrees (0..360)
    'magnitude': [magnitude at_1, ..., magnitude at_N]
    # list of magnitude of the applied fields in Ry units
})
Note: The length of the theta, phi and magnitude lists have to be
equal to the number of atoms.
"""
        )
        spec.input(
            'deciout_parent',
            valid_type=RemoteData,
            required=False,
            help='KkrCalculation RemoteData folder from deci-out calculation'
        )
        spec.input(
            'retrieve_kkrflex',
            valid_type=Bool,
            required=False,
            default=lambda: Bool(True),
            help="""
For a GF writeout calculation, determine whether or not
the kkrflex_* files are copied to the retrieved (can clutter the
database) or are ony left in the remote folder.
"""
        )
        spec.input(
            'anomalous_density',
            valid_type=FolderData,
            required=False,
            help="""
FolderData that contains anomalous density input files for
the KKRhost BdG calculation. If these are not give the code looks
for them in the retrieved of the parent calculation and takes them
from there."""
        )
        spec.input(
            'settings_LDAU',
            valid_type=Dict,
            required=False,
            help="""
Settings for running a LDA+U calculation. The Dict node should be of the form
    settings_LDAU = Dict(dict={'iatom=0':{
        'L': 3,         # l-block which gets U correction (1: p, 2: d, 3: f-electrons)
        'U': 7.,        # U value in eV
        'J': 0.75,      # J value in eV
        'Eref_EF': 0.,  # reference energy in eV relative to the Fermi energy. This is the energy where the projector wavefunctions are calculated (should be close in energy where the states that are shifted lie (e.g. for Eu use the Fermi energy))
    }})
    Note: you can add multiple entries like the one for iatom==0 in this example. The atom index refers to the corresponding atom in the impurity cluster.
"""
        )

        # define outputs
        spec.output(
            'output_parameters',
            valid_type=Dict,
            required=True,
            help='results of the KKR calculation',
        )
        spec.default_output_node = 'output_parameters'

        # define exit codes, also used in parser
        spec.exit_code(
            301,
            'ERROR_NO_OUTPUT_FILE',
            message='KKR output file not found',
        )
        spec.exit_code(
            302,
            'ERROR_KKR_PARSING_FAILED',
            message='KKR parser retuned an error',
        )
        spec.exit_code(
            303,
            'ERROR_NO_SHAPEFUN_FOUND',
            message='Could not find shapefun from voronoi parent',
        )

    def prepare_for_submission(self, tempfolder):
        """
        Create input files.

            :param tempfolder: aiida.common.folders.Folder subclass where
                the plugin should put all its files.
            :param inputdict: dictionary of the input nodes as they would
                be returned by get_inputs_dict
        """

        ########################################
        ### Create input files in tempfolder ###

        # get mandatory input nodes
        parameters = self.inputs.parameters
        code = self.inputs.code
        parent_calc_folder = self.inputs.parent_folder

        # extract parent calculation from input remote folder
        parent_calc = self._get_parent_calc(parent_calc_folder)

        # check if no parameters are overwritten that should not be set manually
        self._check_illegal_parameter_overwrite(parent_calc, parameters)

        # get structure and related inputs from parent voronoi calculation
        voro_parent, structure, vca_structure, use_alat_input = self._get_structure_inputs(parent_calc, parameters)

        # prepare scoef file if impurity_info was given
        parameters = self._write_scoef_file(tempfolder, parameters, structure, use_alat_input)

        # qdos option, ensure low T, E-contour, qdos run option and write qvec.dat file
        if 'kpoints' in self.inputs:
            parameters = self._prepare_qdos_calc(parameters, self.inputs.kpoints, structure, tempfolder, use_alat_input)

        # write nonco_angle.dat file and adapt RUNOPTS if needed (i.e. add FIXMOM if directions are not relaxed)
        if 'initial_noco_angles' in self.inputs:
            parameters = self._use_initial_noco_angles(parameters, structure, tempfolder)

        # write bfield.dat file and add '<NONCOBFIELD>= True' to input parameters
        if 'bfield' in self.inputs:
            parameters = self._use_nonco_bfield(parameters, structure, tempfolder)

        # activate decimation mode and copy decifile from deciout parent
        if 'deciout_parent' in self.inputs:
            parameters = self._use_decimation(parameters, tempfolder)

        # Prepare inputcard from Structure and input parameter data
        shapes = self._get_shapes_array(parent_calc, voro_parent)
        with tempfolder.open(self._INPUT_FILE_NAME, u'w') as input_file:
            natom, nspin, newsosol, warnings_write_inputcard = generate_inputcard_from_structure(
                parameters,
                structure,
                input_file,
                parent_calc,
                shapes=shapes,
                vca_structure=vca_structure,
                use_input_alat=use_alat_input
            )

        #################################################
        ### Copy input files from parent calculations ###

        # add BdG input potential if it is there in the parent calculation output
        self._copy_BdG_pot(parent_calc.outputs.retrieved, tempfolder)

        # get the local copy list (files copied from parent's retrieved to tempfolder at submission)
        local_copy_list = self._get_local_copy_list(parent_calc, voro_parent, tempfolder, parameters)
        # TODO use remote_copy_list or even symlinks if we can copy files between remote working dirs

        #############################################################################
        ### Get list of files that we want to retrieve after the calculation ends ###

        # Retrieve list needs some logic, retrieve certain files,
        # only if certain input keys are specified....
        retrieve_list = self._get_default_retrieve_list()

        # for special cases add files to retireve list:

        # 1. dos calculation, add *dos* files if NPOL==0
        retrieve_list += self._get_dos_filelist(parameters, natom, nspin)

        # 2. KKRFLEX calculation
        retrieve_list += self._get_kkrflex_filelist(parameters)

        # 3. qdos calculation output
        retrieve_list += self._get_qdos_filelist(parameters, natom, nspin)

        # 4. Jij calculation outputs
        retrieve_list += self._get_Jij_filelist(parameters, natom)

        # 5. deci-out (input file for decimation calculation)
        retrieve_list += self._get_deci_file(parameters)

        # 6. BdG output files (anomalous density and (q)dos files with _eh etc. endings)
        retrieve_list += self._get_BdG_filelist(parameters, natom, nspin)

        # 7. retrieve LDA+U potential
        # write ldaupot file and change retrieved list for LDA+U calculation
        # this is triggered with having the settings_LDAU dict node in input.
        retrieve_list += self._init_ldau(tempfolder, parent_calc, natom)

        ####################################################
        ### Collect all inputs and put into the CalcInfo ###

        # Prepare CalcInfo to be returned to aiida
        calcinfo = CalcInfo()
        calcinfo.uuid = self.uuid
        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = []
        calcinfo.retrieve_list = retrieve_list

        # now set calcinfo and return
        codeinfo = CodeInfo()
        codeinfo.cmdline_params = []
        codeinfo.code_uuid = code.uuid
        codeinfo.stdout_name = self._DEFAULT_OUTPUT_FILE
        calcinfo.codes_info = [codeinfo]

        return calcinfo

    def _get_parent_calc(self, parent_calc_folder):
        """Get the parent calculation from the remote folder that is given as input"""

        parent_calcs = parent_calc_folder.get_incoming(node_class=CalcJobNode)

        # check if parent is unique
        n_parents = len(parent_calcs.all_link_labels())
        if n_parents != 1:
            raise UniquenessError(
                'Input RemoteData is child of {} '
                'calculation{}, while it should have a single parent'
                ''.format(n_parents, '' if n_parents == 0 else 's')
            )
            # TODO change to exit code

        parent_calc = parent_calcs.first().node

        return parent_calc

    def _check_illegal_parameter_overwrite(self, parent_calc, parameters):
        """check if no keys are illegally overwritten (i.e. compare with keys in self._do_never_modify)"""

        # extract parent input parameter dict for following check
        try:
            parent_inp_dict = parent_calc.inputs.parameters.get_dict()
        except:
            self.logger.error(f'Failed trying to find input parameter of parent {parent_calc}')
            raise InputValidationError('No parameter node found of parent calculation.')

        # now go though list of parameters which are overwritten and check if that is forbidden
        for key in list(parameters.get_dict().keys()):
            value = parameters.get_dict()[key]
            #self.logger.info("Checking {} {}".format(key, value))
            if value is not None:
                if key in self._do_never_modify:
                    oldvalue = parent_inp_dict[key]
                    try:
                        if oldvalue is None and key in __kkr_default_params__:
                            oldvalue = __kkr_default_params__.get(key)
                        if value == oldvalue:
                            values_eqivalent = True
                        else:
                            values_eqivalent = False
                            # check if values match up to certain numerical accuracy
                            if type(value) == float:
                                if abs(value - oldvalue) < self._eps:
                                    values_eqivalent = True
                            elif type(value) == list or type(value) == np.ndarray:
                                tmp_value, tmp_oldvalue = np.array(value).reshape(-1), np.array(oldvalue).reshape(-1)
                                values_eqivalent_tmp = []
                                for ival in range(len(tmp_value)):
                                    if abs(tmp_value[ival] - tmp_oldvalue[ival]) < self._eps:
                                        values_eqivalent_tmp.append(True)
                                    else:
                                        values_eqivalent_tmp.append(False)
                                if all(values_eqivalent_tmp) and len(value) == len(oldvalue):
                                    values_eqivalent = True
                    except:
                        raise InputValidationError(
                            f'Error while trying to compare old and new values with key={key} in do_never_modify list, oldval={oldvalue}; newval={value}'
                        )
                    if not values_eqivalent:
                        self.logger.error(
                            f'You are trying to set keyword {key} = {value} but this is not allowed since the structure would be modified. Please use a suitable workfunction instead.'
                        )
                        raise InputValidationError(
                            f'You are trying to modify a keyword that is not allowed to be changed! (key={key}, oldvalue={oldvalue}, newvalue={value})'
                        )

    def _get_structure_inputs(self, parent_calc, parameters):
        """extract structure, voronoi parent calculation and some structure properties"""

        # get StructureData node from Parent if Voronoi
        structure = None
        self.logger.info('KkrCalculation: Get structure node from voronoi parent')
        try:
            structure, voro_parent = find_parent_structure(parent_calc)
        except:
            self.logger.error(f'KkrCalculation: Could not get structure from Voronoi parent ({parent_calc}).')
            raise ValidationError(f'Cound not find structure node from parent {parent_calc}')

        # for VCA: check if input structure and parameter node define VCA structure
        vca_structure = vca_check(structure, parameters)

        # check whether or not the alat from the input parameters are used
        # (this enters as a scaling factor for some parameters)
        use_alat_input = parameters.get_dict().get('use_input_alat', False)
        use_alat_input = parameters.get_dict().get('USE_INPUT_ALAT', use_alat_input)

        # Check for 2D case
        twoDimcheck, msg = check_2Dinput_consistency(structure, parameters)
        if not twoDimcheck:
            raise InputValidationError(msg)

        return voro_parent, structure, vca_structure, use_alat_input

    def _write_scoef_file(self, tempfolder, parameters, structure, use_alat_input):
        """Write the scoef file for KKRFLEX writeout"""

        # for GF writeout
        if 'impurity_info' in self.inputs:
            imp_info = self.inputs.impurity_info
            found_imp_info = True
        else:
            imp_info = None
            found_imp_info = False

        write_scoef = False
        runopt = parameters.get_dict().get('RUNOPT', None)

        if found_imp_info:
            self.logger.info('Found impurity_info in inputs of the calculation, automatically add runopt KKRFLEX')
            write_scoef = True
            change_values = []
            if runopt is None:
                runopt = []
            runopt = [i.strip() for i in runopt]
            if 'KKRFLEX' not in runopt:
                runopt.append('KKRFLEX')
                change_values.append(['RUNOPT', runopt])
            parameters = _update_params(parameters, change_values)
        elif runopt is not None and 'KKRFLEX' in runopt:
            # if we end up here there is a problem with the input
            self.logger.info('Need to write scoef file but no impurity_info given!')
            raise ValidationError('Found RUNOPT KKRFLEX but no impurity_info in inputs')

        if found_imp_info and write_scoef:

            imp_info_dict = imp_info.get_dict()

            # find alat input if needed
            if use_alat_input:
                alat_input = parameters.get_dict().get('ALATBASIS', None) / get_Ang2aBohr()
                self.logger.info('alat_input is ' + str(alat_input))
            else:
                self.logger.info('alat_input is None')
                alat_input = None

            # create scoef file
            if 'imp_cls' not in imp_info_dict:
                # this means cluster is found from parameters in imp_info

                # extract cluster settings
                Rcut = imp_info_dict.get('Rcut', None)
                hcut = imp_info_dict.get('hcut', -1.)
                cylinder_orient = imp_info_dict.get('cylinder_orient', [0., 0., 1.])
                ilayer_center = imp_info_dict.get('ilayer_center', 0)
                for i in range(len(cylinder_orient)):
                    try:
                        len(cylinder_orient[i])
                        vec_shape = False
                    except TypeError:
                        vec_shape = True

                # some consistency checks
                if ilayer_center > len(structure.sites) - 1:
                    raise IndexError(
                        f'Index of the reference site is out of range! Possible values: 0 to {len(structure.sites) - 1}.'
                    )
                elif Rcut < 0:
                    raise ValueError('Cutoff radius has to be positive!')
                if not vec_shape or len(cylinder_orient) != 3:
                    raise TypeError(
                        f'Input orientation vector ({cylinder_orient}) has the wrong shape! It needs to be a 3D-vector!'
                    )

                # now write scoef file
                with tempfolder.open(self._SCOEF, 'w') as scoef_file:
                    make_scoef(
                        structure,
                        Rcut,
                        scoef_file,
                        hcut,
                        cylinder_orient,
                        ilayer_center,
                        alat_input,
                    )

            else:

                # this means the full imp cluster is given in the input
                # TODO add some consistency checks with structure etc.
                with tempfolder.open(self._SCOEF, 'w') as scoef_file:
                    if alat_input is not None:
                        alat = get_alat_from_bravais(np.array(structure.cell), structure.pbc[2])
                        rescale_alat = alat / alat_input
                        self.report(f'INFO: rescaling imp cls due to alat_input: {rescale_alat}')
                    else:
                        rescale_alat = None
                    write_scoef_full_imp_cls(imp_info, scoef_file, rescale_alat)

        return parameters

    def _get_shapes_array(self, parent_calc, voro_parent):
        """Get the shapes array from the parent calcualtion or the voronoi parent"""

        # set shapes array either from parent voronoi run
        if parent_calc.process_label == 'VoronoiCalculation' or parent_calc.process_label == 'KkrCalculation':
            # get shapes array from voronoi parent
            shapes = voro_parent.outputs.output_parameters.get_dict().get('shapes')
        else:
            # extract shapes from input parameters node constructed by kkrimporter calculation
            shapes = voro_parent.inputs.parameters.get_dict().get('<SHAPE>')

        self.logger.info(f'Extracted shapes: {shapes}')

        return shapes

    def _get_local_copy_list(self, parent_calc, voro_parent, tempfolder, parameters):
        """Decide what files to copy based on settings to the code (e.g. KKRFLEX option needs scoef)"""
        # copy the right files #TODO check first if file, exists and throw
        # warning, now this will throw an error
        outfolder = parent_calc.outputs.retrieved

        copylist = []
        if parent_calc.process_class == KkrCalculation:
            copylist = [self._OUT_POTENTIAL]
            # TODO ggf copy remotely from remote node if present ...
        elif parent_calc.process_class == VoronoiCalculation:
            copylist = [parent_calc.process_class._SHAPEFUN]
            # copy either overwrite potential or voronoi output potential
            # (voronoi caclualtion retreives only one of the two)
            if parent_calc.process_class._POTENTIAL_IN_OVERWRITE in outfolder.list_object_names():
                copylist.append(parent_calc.process_class._POTENTIAL_IN_OVERWRITE)
            else:
                copylist.append(parent_calc.process_class._OUT_POTENTIAL_voronoi)
        else:  #  (if parent_calc.process_class == KkrImporterCalculation:)
            #change copylist in case the calculation starts from an imported calculation
            if self._OUT_POTENTIAL in outfolder.list_object_names():
                copylist.append(self._OUT_POTENTIAL)
            else:
                copylist.append(self._POTENTIAL)
            if self._SHAPEFUN in outfolder.list_object_names():
                copylist.append(self._SHAPEFUN)

        # create local_copy_list from copylist and change some names automatically
        local_copy_list = []
        for file1 in copylist:
            # deal with special case that file is written to another name
            if (
                file1 == 'output.pot' or file1 == self._OUT_POTENTIAL or (
                    parent_calc.process_class == VoronoiCalculation and
                    file1 == parent_calc.process_class._POTENTIAL_IN_OVERWRITE
                )
            ):
                filename = self._POTENTIAL
            else:
                filename = file1
            # now add to copy list
            local_copy_list.append((outfolder.uuid, file1, filename))

            # add shapefun file from voronoi parent if needed
            if self._SHAPEFUN not in copylist:
                try:
                    struc, voro_parent = find_parent_structure(parent_calc)
                except ValueError:
                    return self.exit_codes.ERROR_NO_SHAPEFUN_FOUND  # pylint: disable=no-member
                # copy shapefun from retrieved of voro calc
                voro_retrieved = voro_parent.outputs.retrieved
                local_copy_list.append((voro_retrieved.uuid, VoronoiCalculation._SHAPEFUN, self._SHAPEFUN))

        # check if core state lie within energy contour and take them out if needed
        local_copy_list = self._kick_out_corestates_kkrhost(local_copy_list, tempfolder)

        # for set-ef option (needs to be done AFTER kicking out core states):
        ef_set = parameters.get_dict().get('ef_set', None)
        ef_set = parameters.get_dict().get('EF_SET', ef_set)
        if verbose:
            set_values = kkrparams(**parameters.get_dict()).get_set_values()
            self.report(
                f"efset: {ef_set}  efset_1: {parameters.get_dict().get('ef_set')} efset_2: {parameters.get_dict().get('EF_SET')} params: {set_values}"
            )
        if ef_set is not None:
            local_copy_list = self._set_ef_value_potential(ef_set, local_copy_list, tempfolder)

        # TODO different copy lists, depending on the keywors input
        self.logger.info(f'local copy list: {local_copy_list}')

        return local_copy_list

    def _get_default_retrieve_list(self):
        """initialize retrieve list with these files"""

        ret_list = [
            self._DEFAULT_OUTPUT_FILE,
            self._INPUT_FILE_NAME,
            self._SCOEF,
            self._NONCO_ANGLES_OUT,
            self._NONCO_ANGLES_ALL_ITER,
            self._OUT_POTENTIAL,
            self._OUTPUT_0_INIT,
            self._OUTPUT_000,
            self._OUTPUT_2,
            self._OUT_TIMING_000,
        ]

        return ret_list

    def _get_dos_filelist(self, parameters, natom, nspin, addition=''):
        """return file list of DOS output files to add to retrieve list """

        retrieve_dos_files = False
        add_files = []

        if 'NPOL' in list(parameters.get_dict().keys()):
            if parameters.get_dict()['NPOL'] == 0:
                retrieve_dos_files = True

        if 'TESTOPT' in list(parameters.get_dict().keys()):
            testopts = parameters.get_dict()['TESTOPT']
            if testopts is not None:
                stripped_test_opts = [i.strip() for i in testopts]
                if 'DOS' in stripped_test_opts:
                    retrieve_dos_files = True

        if retrieve_dos_files:
            if verbose:
                self.report(f'adding files for dos output: {[self._COMPLEXDOS, self._DOS_ATOM, self._LMDOS]}')

            add_files = [self._COMPLEXDOS + addition]
            for iatom in range(natom):
                add_files.append(self._DOS_ATOM % (iatom + 1) + addition)
                for ispin in range(nspin):
                    add_files.append((self._LMDOS % (iatom + 1, ispin + 1)).replace(' ', '0') + addition)

        return add_files

    def _get_kkrflex_filelist(self, parameters):
        """Add kkrflex_* files to the retrieve list if needed"""

        retrieve_kkrflex_files = False
        add_files = []

        if 'RUNOPT' in list(parameters.get_dict().keys()):
            runopts = parameters.get_dict()['RUNOPT']
            if runopts is not None:
                stripped_run_opts = [i.strip() for i in runopts]
                if 'KKRFLEX' in stripped_run_opts:
                    retrieve_kkrflex_files = True

        if retrieve_kkrflex_files:
            if 'retrieve_kkrflex' in self.inputs and self.inputs.retrieve_kkrflex.value:
                # retrieve all kkrflex files
                add_files = self._ALL_KKRFLEX_FILES
            else:
                # do not retrieve kkrflex_tmat and kkrflex_green, they are kept on the remote and used from there
                add_files = [self._KKRFLEX_ATOMINFO, self._KKRFLEX_INTERCELL_REF, self._KKRFLEX_INTERCELL_CMOMS]

        return add_files

    def _get_qdos_filelist(self, parameters, natom, nspin, addition=''):
        """Add qdos output files to retrieve list if needed"""

        retrieve_qdos_files = False
        add_files = []

        if 'RUNOPT' in list(parameters.get_dict().keys()):
            runopts = parameters.get_dict()['RUNOPT']
            if runopts is not None:
                stripped_run_opts = [i.strip() for i in runopts]
                if 'qdos' in stripped_run_opts:
                    retrieve_qdos_files = True

        if retrieve_qdos_files:
            add_files = [self._QVEC]
            for iatom in range(natom):
                for ispin in range(nspin):
                    add_files.append((self._QDOS_ATOM % (iatom + 1, ispin + 1)).replace(' ', '0') + addition)
                    # try to retrieve both old and new version of the files
                    add_files.append((self._QDOS_ATOM_OLD % (iatom + 1, ispin + 1)).replace(' ', '0') + addition)
                # retrieve also qdos_sx,y,z files if written out
                add_files.append((self._QDOS_SX % (iatom + 1)).replace(' ', '0') + addition)
                add_files.append((self._QDOS_SY % (iatom + 1)).replace(' ', '0') + addition)
                add_files.append((self._QDOS_SZ % (iatom + 1)).replace(' ', '0') + addition)

        return add_files

    def _get_Jij_filelist(self, parameters, natom):
        """Add Jij output files to retrieve list"""
        retrieve_Jij_files = False
        add_files = []

        if 'RUNOPT' in list(parameters.get_dict().keys()):
            runopts = parameters.get_dict()['RUNOPT']
            if runopts is not None:
                stripped_run_opts = [i.strip() for i in runopts]
                if 'XCPL' in stripped_run_opts:
                    retrieve_Jij_files = True

        if retrieve_Jij_files:
            add_files = [self._SHELLS_DAT] + [self._Jij_ATOM % iatom for iatom in range(1, natom + 1)]

        return add_files

    def _get_deci_file(self, parameters):
        """Add deci-out file to retrieve list"""

        retrieve_decifile = False
        add_files = []

        if 'RUNOPT' in list(parameters.get_dict().keys()):
            runopts = parameters.get_dict()['RUNOPT']
            if runopts is not None:
                stripped_run_opts = [i.strip() for i in runopts]
                if 'deci-out' in stripped_run_opts:
                    retrieve_decifile = True

        if retrieve_decifile:
            add_files = [self._DECIFILE]

        return add_files

    def _get_BdG_filelist(self, parameters, natom, nspin):
        """Add list of BdG output files to retrieve list"""

        retrieve_BdG_files = False
        add_files = []

        use_BdG_dict = {
            k.upper().replace('<', '').replace('>', ''): v
            for k, v in parameters.get_dict().items()
            if 'USE_BDG' == k.upper().replace('<', '').replace('>', '')
        }
        retrieve_BdG_files = use_BdG_dict.get('USE_BDG', False)

        if verbose:
            self.report(f'retrieve BdG? {retrieve_BdG_files}')

        if retrieve_BdG_files:
            # anomalous density files for all atoms if present
            for iatom in range(natom):
                for ispin in range(nspin):
                    print('adding files for BdG mode')
                    add_files += [self._BDG_POT % (iatom + 1, ispin + 1)]
            # radially-averaged anomalous density matrix (for triplet components etc.)
            add_files.append(self._BDG_CHI_NS)

            #also retrieve BdG DOS files for anomalous density and hole part
            for BdGadd in ['_eh', '_he', '_hole']:
                add_files += self._get_dos_filelist(parameters, natom, nspin, BdGadd)
                add_files += self._get_qdos_filelist(parameters, natom, nspin, BdGadd)

        return add_files

    def _set_parent_remotedata(self, remotedata):
        """
        Used to set a parent remotefolder in the restart of fleur.
        """
        if not isinstance(remotedata, RemoteData):
            raise ValueError('remotedata must be a RemoteData')

        # complain if another remotedata is already found
        input_remote = self.get_inputs(node_type=RemoteData)
        if input_remote:
            raise ValidationError('Cannot set several parent calculation to a KKR calculation')

        self.use_parent_folder(remotedata)

    def _set_ef_value_potential(self, ef_set, local_copy_list, tempfolder):
        """
        Set EF value ef_set in the potential file.
        """
        self.report(f'local copy list before change: {local_copy_list}')
        self.report("found 'ef_set' in parameters: change EF of potential to this value")

        # first read old potential

        if self._POTENTIAL in tempfolder.get_content_list():
            has_potfile = True
        else:
            has_potfile = False
        self.report(f'has_potfile? {has_potfile}')

        txt = []
        if has_potfile:
            # this is the case if we kicked out core states before
            with tempfolder.open(self._POTENTIAL, 'r') as potfile:
                # read potential
                txt = potfile.readlines()

        if not has_potfile or len(txt) == 0:
            # this is the case when we take the potential from an existing folder
            potcopy_info = [i for i in local_copy_list if i[2] == self._POTENTIAL][0]
            with load_node(potcopy_info[0]).open(potcopy_info[1]) as potfile:
                # remove previous output potential from copy list
                local_copy_list.remove(potcopy_info)
                # read potential
                txt = potfile.readlines()

        self.report(f'len(potfile)? {len(txt)}')

        # now change value of Fermi level in potential text
        potstart = []
        for iline in range(len(txt)):
            line = txt[iline]
            if 'exc:' in line:
                potstart.append(iline)
        for ipotstart in potstart:
            self.report(f'set ef {ef_set} in potential starting in line {ipotstart}')
            tmpline = txt[ipotstart + 3]
            tmpline = tmpline.split()
            newline = f'{float(tmpline[0]):10.5f}{ef_set:20.14f}{float(tmpline[-1]):20.14f}\n'

            txt[ipotstart + 3] = newline

        # now (over)writing potential file in tempfolder with changed Fermi energy
        with tempfolder.open(self._POTENTIAL, 'w') as pot_new_ef:
            # write new file
            pot_new_ef.writelines(txt)
            # now this directory (tempfolder) contains the updated potential file
            # thus it is not needed to put it in the local copy list anymore

        # return updated local_copy_list
        return local_copy_list

    def _kick_out_corestates_kkrhost(self, local_copy_list, tempfolder):
        """
        Compare value of core states from potential file in local_copy_list with EMIN
        and kick corestate out of potential if they lie inside the energy contour.
        """

        # read EMIN value from inputcard
        params = kkrparams()
        with tempfolder.open(self._INPUT_FILE_NAME) as input_file:
            params.read_keywords_from_inputcard(input_file)
        emin = params.get_value('EMIN')

        # run kick_out_corestates routine to remove core states that lie above emin
        potcopy_info = [i for i in local_copy_list if i[2] == self._POTENTIAL][0]
        with tempfolder.open(self._POTENTIAL, 'w') as potfile_out:
            with load_node(potcopy_info[0]).open(potcopy_info[1]) as potfile_in:
                num_deleted = kick_out_corestates(potfile_in, potfile_out, emin)

        # remove changed potential from local copy list (already in tempfolder without overlapping core states)
        if num_deleted > 0:
            local_copy_list.remove(potcopy_info)
        else:
            # remove temporarily created file
            tempfolder.remove_path(self._POTENTIAL)

        # return updated local_copy_list
        return local_copy_list

    def _prepare_qdos_calc(self, parameters, kpath, structure, tempfolder, use_alat_input):
        """
        prepare a qdos (i.e. bandstructure) calculation, can only be done if k-points are given in input
        Note: this changes some settings in the parameters to ensure a DOS contour and low smearing temperature
        Also the qvec.dat file is written here.
        """
        # check qdos settings
        change_values = []
        runopt = parameters.get_dict().get('RUNOPT')
        if runopt is None:
            runopt = []
        runopt = [i.strip() for i in runopt]
        if 'qdos' not in runopt:
            runopt.append('qdos')
            change_values.append(['RUNOPT', runopt])
        tempr = parameters.get_dict().get('TEMPR')
        if tempr is None or tempr > 100.:
            change_values.append(['TEMPR', 50.])
        N1 = parameters.get_dict().get('NPT1')
        if N1 is None or N1 > 0:
            change_values.append(['NPT1', 0])
        N2 = parameters.get_dict().get('NPT2')
        if N2 is None:
            change_values.append(['NPT2', 100])
        N3 = parameters.get_dict().get('NPT3')
        if N3 is None or N3 > 0.:
            change_values.append(['NPT3', 0])
        NPOL = parameters.get_dict().get('NPOL')
        if NPOL is None or NPOL > 0.:
            change_values.append(['NPOL', 0])
        parameters = _update_params(parameters, change_values)
        # write qvec.dat file
        kpath_array = kpath.get_kpoints(cartesian=True)
        # convert automatically to internal units
        alat = get_alat_from_bravais(np.array(structure.cell), is3D=structure.pbc[2]) * get_Ang2aBohr()
        if use_alat_input:
            alat_input = parameters.get_dict().get('ALATBASIS')
        else:
            alat_input = alat
        kpath_array = kpath_array * (alat_input / alat) / get_Ang2aBohr() / (2 * np.pi / alat)
        # now write file
        qvec = [f'{len(kpath_array)}\n']
        qvec += [f'{kpt[0]:e} {kpt[1]:e} {kpt[2]:e}\n' for kpt in kpath_array]
        with tempfolder.open(self._QVEC, 'w') as qvecfile:
            qvecfile.writelines(qvec)

        return parameters

    def _use_initial_noco_angles(self, parameters, structure, tempfolder):
        """
        Set starting values for non-collinear calculation (writes nonco_angle.dat to tempfolder).
        Adapt FIXMOM runopt according to fix_dir input in initial_noco_angle input node
        """
        self.report('Found `initial_noco_angles` input node, writing nonco_angle.dat file')

        # extract fix_dir flag and set FIXMOM RUNOPT in parameters accordingly
        fix_dir = self.inputs.initial_noco_angles['fix_dir']
        natom = get_natyp(structure)
        if len(fix_dir) != natom:
            raise InputValidationError(
                'Error: `fix_dir` list in `initial_noco_angles` input node needs to have the same length as number of atoms!'
            )

        change_values = []
        runopt = parameters.get_dict().get('RUNOPT')
        if runopt is None:
            runopt = []
        runopt = [i.strip() for i in runopt]
        if all(fix_dir) and 'FIXMOM' not in runopt:
            runopt.append('FIXMOM')
            change_values.append(['RUNOPT', runopt])
        elif not all(fix_dir) and 'FIXMOM' in [i.upper() for i in runopt]:
            runopt = [i for i in runopt if 'FIXMOM' not in i.upper()]
            change_values.append(['RUNOPT', runopt])
        parameters = _update_params(parameters, change_values)

        # extract theta and phi values from input node
        thetas = self.inputs.initial_noco_angles['theta']
        if len(thetas) != natom:
            raise InputValidationError(
                'Error: `theta` list in `initial_noco_angles` input node needs to have the same length as number of atoms!'
            )
        phis = self.inputs.initial_noco_angles['phi']
        if len(phis) != natom:
            raise InputValidationError(
                'Error: `phi` list in `initial_noco_angles` input node needs to have the same length as number of atoms!'
            )

        # now write kkrflex_angle file
        with tempfolder.open(self._NONCO_ANGLES, 'w') as noco_angle_file:
            for iatom in range(natom):
                theta, phi = thetas[iatom], phis[iatom]
                # check consistency
                if theta < 0. or theta > 180.:
                    raise InputValidationError(
                        f'Error: theta value out of range (0..180): iatom={iatom}, theta={theta}'
                    )
                # convert fix_dir to boolean if given as integer
                if not isinstance(fix_dir[iatom], bool):
                    fix_dir[iatom] = (fix_dir[iatom] == 1)
                # write line
                noco_angle_file.write(f'   {theta}    {phi}    {fix_dir[iatom]}\n')

        return parameters

    def _use_nonco_bfield(self, parameters, structure, tempfolder):
        """
        Set external non-collinear bfield (writes bfield.dat to tempfolder) used in constraint calculations.
        """
        self.report('Found `bfield` input node, writing nonco_angle.dat file')

        # extract number of atoms for length comparison
        natom = get_natyp(structure)

        change_values = [['<NONCOBFIELD>', True]]
        parameters = _update_params(parameters, change_values)

        # extract magnitude, theta and phi values from input node
        mags = self.inputs.bfield['magnitude']
        if len(mags) != natom:
            raise InputValidationError(
                'Error: `magnitude` list in `bfield` input node needs to have the same length as number of atoms!'
            )
        thetas = self.inputs.bfield['theta']
        if len(thetas) != natom:
            raise InputValidationError(
                'Error: `theta` list in `bfield` input node needs to have the same length as number of atoms!'
            )
        phis = self.inputs.bfield['phi']
        if len(phis) != natom:
            raise InputValidationError(
                'Error: `phi` list in `bfield` input node needs to have the same length as number of atoms!'
            )

        # now write kkrflex_angle file
        with tempfolder.open(self._BFIELD, 'w') as bfield_file:
            bfield_file.write('# theta [deg]  phi [deg]  magnitude [Ry]\n')
            for iatom in range(natom):
                theta, phi = thetas[iatom], phis[iatom]
                magnitude = mags[iatom]
                # check consistency
                if theta < 0. or theta > 180.:
                    raise InputValidationError(
                        f'Error: theta value out of range (0..180): iatom={iatom}, theta={theta}'
                    )
                # write line
                bfield_file.write(f'   {theta}    {phi}    {magnitude}\n')

        return parameters

    def _use_decimation(self, parameters, tempfolder):
        """
        Activate decimation mode and copy decifile from output of deciout_parent calculation
        """
        self.report('Found `deciout_parent` input node, activae decimation mode')

        # check if deciout parent calculation was proper deci-out calculation
        deciout_parent = self.inputs.deciout_parent
        parent_calcs = deciout_parent.get_incoming(node_class=CalcJobNode)
        n_parents = len(parent_calcs.all_link_labels())
        if n_parents != 1:
            raise UniquenessError(
                'Input RemoteData is child of {} '
                'calculation{}, while it should have a single parent'
                ''.format(n_parents, '' if n_parents == 0 else 's')
            )
            # TODO change to exit code
        parent_calc = parent_calcs.first().node
        deciout_retrieved = parent_calc.outputs.retrieved
        if self._DECIFILE not in deciout_retrieved.list_object_names():
            raise InputValidationError('Error: deciout_parent does not contain decifile!')

        # add 'DECIMATE' flag, decifile and NSTEPS=1
        change_values = []
        runopt = parameters.get_dict().get('RUNOPT')
        if runopt is None:
            runopt = []
        runopt = [i.strip() for i in runopt]
        runopt.append('DECIMATE')
        change_values.append(['RUNOPT', runopt])
        change_values.append(['FILES', [self._POTENTIAL, self._SHAPEFUN]])  # needed to make DECIFILE work
        change_values.append(['DECIFILES', ['vacuum', self._DECIFILE]])  # works only for right continuation for now!
        change_values.append(['NSTEPS', 1])  # decimation works only in one-shot mode
        parameters = _update_params(parameters, change_values)

        # now write kkrflex_angle file
        with deciout_retrieved.open(self._DECIFILE, 'r') as decifile_handle:
            decifile_txt = decifile_handle.readlines()
        with tempfolder.open(self._DECIFILE, 'w') as decifile_handle:
            decifile_handle.writelines(decifile_txt)

        return parameters

    def _copy_BdG_pot(self, retrieved, tempfolder):
        """
        Activate BdG mode and copy den_lm_ir files of the previous output to the input of this calculation.
        """

        if 'anomalous_density' in self.inputs:
            # this means we have a FolderData input that contains the
            # anomalous density files
            adens_folder = self.inputs.anomalous_density
        else:
            # if no anomalous density is given as an input node we check
            # if there are any anomalous density files in the parent retrieved
            # and take them from there if present
            adens_folder = retrieved

        # list of den_lm_ir files (anomalous density per atom)
        BDG_POT_FILES = [i for i in adens_folder.list_object_names() if self._BDG_POT.split('.')[0] in i]

        # add 'den-lm_ir' files to input
        for BdG_pot in BDG_POT_FILES:
            self.report(f'Copy BdG potential {BdG_pot} from {adens_folder.uuid}')
            # read from parent
            with adens_folder.open(BdG_pot, 'r') as file_handle:
                file_txt = file_handle.readlines()
            # write to tempfolder
            with tempfolder.open(BdG_pot, 'w') as file_handle:
                file_handle.writelines(file_txt)

    def _init_ldau(self, tempfolder, parent_calc, natom):
        """
        Check if settings_LDAU is in input and set up LDA+U calculation. Reuse old ldaupot of parent_folder contains a file ldaupot.
        """
        retrieve_list = []

        # first check if settings_LDAU is in inputs
        if 'settings_LDAU' not in self.inputs:
            # do nothing
            return retrieve_list

        else:
            # this means we need to set up LDA+U

            # add ldaupot to retrieve and local copy lists
            retrieve_list.append(self._LDAUPOT + '_new')

            # add runoption for LDA+U
            with tempfolder.open(self._INPUT_FILE_NAME) as file:
                inputcard = file.readlines()
            itmp = search_string('RUNOPT', inputcard)

            if itmp >= 0:
                # add LDA+U as runoption
                inputcard[itmp + 1] = 'LDA+U   ' + inputcard[itmp + 1]
            else:
                # activate LDA+U with keyword
                inputcard.append('\n# activate LDA+U\n<USE_LDAU>= True\n')

            # create ldaupot file
            reuse_old_ldaupot, lopt, jeff, ueff, eref, atyp = self._create_or_update_ldaupot(
                parent_calc, tempfolder, natom
            )

            # set LDA+U inputs in inputcard
            nat_ldau = len(lopt)
            inputcard.append(f'NAT_LDAU= {nat_ldau}\n')
            inputcard.append(f'LDAU_PARA\n')
            for iat_ldau in range(nat_ldau):
                inputcard.append(
                    f'{atyp[iat_ldau]+1} {lopt[iat_ldau]} {ueff[iat_ldau]:16.7e} {jeff[iat_ldau]:16.7e} {eref[iat_ldau]:16.7e}\n'
                )

            if reuse_old_ldaupot:
                inputcard.append('KREADLDAU= 1\n')
            else:
                inputcard.append('KREADLDAU= 0\n')

            # overwrite inputcard
            with tempfolder.open(self._INPUT_FILE_NAME, 'w') as file:
                file.writelines(inputcard)

        return retrieve_list

    def _create_or_update_ldaupot(self, parent_calc, tempfolder, natom):
        """
        Writes ldaupot to tempfolder.

        If parent_calc is found and it contains an onld ldaupot, we reuse the values for wldau, uldau and phi from there.
        """

        # extract Fermi energy from parent calculation
        ef_Ry = parent_calc.outputs.output_parameters['fermi_energy']

        # get old ldaupot file
        reuse_old_ldaupot = self._get_old_ldaupot(parent_calc, tempfolder)

        # settings dict (defines U, J etc.)
        ldau_settings = self.inputs.settings_LDAU.get_dict()

        if reuse_old_ldaupot:
            # reuse wldau, ildau and phi from old ldaupot file
            # Attention the first number needs to be non-zero
            txt, lopt, jeff, ueff, eref, atyp = get_ldaupot_text(
                ldau_settings, ef_Ry, natom, initialize=False, return_luj=True
            )
            # now we read the old file
            with tempfolder.open(self._LDAUPOT + '_old', 'r') as ldaupot_file:
                txt0 = ldaupot_file.readlines()
                # find start of wldau etc.
                ii = 0
                for line in txt0:
                    if 'wldau' in line.lower():
                        break
                    ii += 1
                txt0 = txt0[ii:]

            # create header for ldaupot file
            txt = txt[0:2] + [f'{len(atyp)}\nLOPT 1.. {len(atyp)}\n']
            txt += [f'  {l}' for l in lopt] + ['\n']
            txt += ['IAT  UEFF  JEFF  EREF\n']
            for ii, iat in enumerate(atyp):
                txt += [f'{iat+1} {ueff[ii]:16.9e} {jeff[ii]:16.9e} {eref[ii]:16.9e}\n']

            # put new header and old bottom together
            newtxt = txt + txt0
        else:
            # initialize ldaupot file
            # Attention: here the first number needs to be 0 which triggers generating initial values in KKRimp
            newtxt, lopt, jeff, ueff, eref, atyp = get_ldaupot_text(
                ldau_settings, ef_Ry, natom, initialize=True, return_luj=True
            )

        # now write to file
        with tempfolder.open(self._LDAUPOT, 'w') as out_filehandle:
            out_filehandle.writelines(newtxt)

        return reuse_old_ldaupot, lopt, jeff, ueff, eref, atyp

    def _get_old_ldaupot(self, parent_calc, tempfolder):
        """
        Copy old ldaupot from retrieved of parent or extract from tarball.
        If no parent_calc is present this step is skipped.
        """

        has_ldaupot = False

        if parent_calc is not None:
            retrieved = parent_calc.outputs.retrieved
            # copy old file to tempfolder
            if self._LDAUPOT + '_new' in retrieved.list_object_names():
                has_ldaupot = True
                with tempfolder.open(self._LDAUPOT + '_old', u'w') as newfile:
                    with retrieved.open(self._LDAUPOT + '_new', u'r') as oldfile:
                        newfile.writelines(oldfile.readlines())

        return has_ldaupot


def _update_params(parameters, change_values):
    """
    change parameters node from change_values list of key value pairs
    Retrun input parameter node if change_values list is empty
    """
    if change_values != []:
        new_params = {}
        for key, val in parameters.get_dict().items():
            new_params[key] = val
        for key, val in change_values:
            new_params[key] = val

        new_params_node = Dict(new_params)

        parameters = new_params_node
    return parameters



"""
these are calcfunctions used to store nodes which ensures the data provenance
"""

from aiida.engine import calcfunction

__copyright__ = (u'Copyright (c), 2020, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.1.0'
__contributors__ = (u'Philipp Ruessmann')


@calcfunction
def create_out_dict_node(out_node, **input_nodes):
    """
    This calcfunction makes sure the data provenance is kept for the output nodes of workflows.
    Stores the output node and links to inputs nodes if given.

    :param out_node: output Dict node containing results of a workflow
    :type out_node: :class: aiida.orm.node

    :param input_nodes: dict of link_label, input_node pairs: `input_nodes = {'link_label', Node}`
    :type node_dict: dict
    ..note::
        The nodes in input_nodes cannot be stored already

    :return output_node: output Node (stored via calcfunction for data-provenance)
    :type output_node: :class: aiida.orm.node
    ..note::
        The output node is a clone of the input `out_node` and will be doubled in the database
    """

    # need to return a clone of the node to not have a cyclic graph
    return out_node.clone()



"""
Parsers provided by the aiida-kkr plugin

Register parsers via the "aiida.parsers" entry point in setup.json.
"""
# import all calculations here to expose them in `aiida_kkr.parsers` directly
from .voro import VoronoiParser
from .kkr import KkrParser
from .kkrimp import KkrimpParser
# broken at the moment
#from .kkrimporter import KkrImporterParser



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Dec 13 14:38:59 2021

@author: markus
"""

# -*- coding: utf-8 -*-

from aiida.parsers.parser import Parser
from aiida.orm import Dict, CalcJobNode
from aiida_kkr.calculations.kkrnano import KKRnanoCalculation
from masci_tools.io.common_functions import (search_string, open_general)
import numpy as np
from io import StringIO
from pprint import pprint as pp
import os

__copyright__ = (u'Copyright (c), 2021, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.0.2'
__contributors__ = ('Markus Struckmann', 'Philipp Rüßmann')


class KKRnanoParser(Parser):
    """
    Parser class for parsing output of the KKRnano code
    """

    def __init__(self, calc):
        """
        Initialize
        """
        # these files should be present after successful run of KKRnano
        #         self._default_files = {
        #             #'stdout': KKRnanoCalculation._DEFAULT_OUTPUT_FILE
        #            #'stdout_prep': KKRnanoCalculation._DEFAULT_OUTPUT_PREP_FILE
        #         }

        self._ParserVersion = __version__

        # reuse init of base class
        super(KKRnanoParser, self).__init__(calc)

    # pylint: disable=protected-access

    def _get_lines(self, retrieved_folder, output_file_name):
        """returns list of string lines"""
        with retrieved_folder.open(output_file_name, 'r') as f:
            lines = f.readlines()
        return lines

    def _findSimpleEntries(self, string2find, retrieved_folder, output_file_name, lineindices=[-1], simpleEntry=True):
        """
        read out entries that are simply given at the end of a line preceeded by the string2find.
        returns a list of said entries
        """
        lines = self._get_lines(retrieved_folder, output_file_name)

        returnlist, indexlist = [], []

        #checking if line indices were passed. If so, only the specified lines are searched
        if lineindices[0] == -1:
            lineindices = range(len(lines))

        #looping over indicated or all (default) lines
        for j in lineindices:
            line = lines[j]
            stringposition = line.find(string2find)

            #matching sought string
            if stringposition >= 0:
                indexlist.append(j)
                keyvalue = line[stringposition + len(string2find) + 1:-1]  #.replace(" ", "")
                keyvalue = keyvalue.replace('=', '')
                keyvalue = keyvalue.split(sep='(')[0]

                if simpleEntry:
                    array = np.genfromtxt(StringIO(keyvalue.replace('D', 'e')), delimiter=' ', dtype=None)  #[0]

                try:
                    returnlist.append(array.item()[0])  # pylint: disable=possibly-used-before-assignment
                except TypeError:
                    returnlist.append(array.item())

        return returnlist, indexlist

    def _get_index_list(self, string2find, retrieved_folder, output_file_name='', lines=[]):
        """
        returns list of indicies of lines containting the passed string.
        opens file, if a file name and no string is passed.
        """

        if lines == [] and output_file_name == '':
            print('ERROR in get_index_list: Neither file name nor lines of strings were passed')
        if lines == []:
            lines = self._get_lines(retrieved_folder, output_file_name)

        returnlist, indexlist = [], []

        for j in range(len(lines)):
            line = lines[j]
            stringposition = line.find(string2find)
            if stringposition >= 0:
                indexlist.append(j)
        return indexlist

    def _read_table_block(self, lines, retrieved_folder, output0_file_name, index_multiple_tables=-1):
        """
        reads a table in the output of the --prepare step which is indicated by a borderlines of "---", returns a string array of said table.
        """
        indexlist = self._get_index_list('---------', retrieved_folder, output0_file_name, lines=lines)
        if index_multiple_tables == -1:
            table = lines[indexlist[-2] + 1:indexlist[-1]]
        else:
            table = lines[indexlist[index_multiple_tables * 2] + 1:indexlist[index_multiple_tables * 2 + 1]]
        array = [line.replace('\n', '').split(' ') for line in table]
        for j in range(len(array)):
            for k in range(array[j].count('')):
                array[j].remove('')

        #array=[np.delete(a,np.where(a=="")) for a in array]
        return np.array(array)

    def _find_block(self, lines, string, retrieved_folder, output0_file_name):
        """
        finds a block which contains the indicated string in passed lines.
        Can be used for the output of the --prepare step which is indicated by a borderlines of "===", returns a list of the lineindices in the passed lines.
        """
        pos__output = search_string(string, lines)
        indexlist = self._get_index_list('=======', retrieved_folder, output0_file_name, lines=lines)
        for i in range(len(indexlist) - 1):
            if indexlist[i] < pos__output and indexlist[i + 1] > pos__output:
                return lines[indexlist[i]:indexlist[i + 1]]
        print(f"Warning: Block '{string}' not found!")
        return []

    def _get_total_EnergyeV(self, key, retrieved_folder, output_file_name, stringsInOutputFile):
        return_dict = {}
        _, indexlist = self._findSimpleEntries(stringsInOutputFile[key], retrieved_folder, output_file_name)
        #print(findSimpleEntries("eV  :",output_file_name,retrieved_folder) np.array(indexlist)+1))
        return_dict['total_energy_in_eV'], _ = self._findSimpleEntries(
            'eV  :', retrieved_folder, output_file_name,
            np.array(indexlist) + 1
        )
        return return_dict

    def _get_charge_in_WScell(self, key, retrieved_folder, output_file_name, n_atoms):
        number_of_atoms = 5
        charges, indexlist = self._findSimpleEntries(key, retrieved_folder, output_file_name)
        charge_dict = {}
        charge_dict['atom'] = {}
        num_atoms = n_atoms

        for k in range(num_atoms):
            charge_dict['atom'][k + 1] = [charges[i] for i in np.arange(0 + k, len(charges) + k, num_atoms)]
        return charge_dict

    def _stringFromList(self, stringlist):
        """
        turn a list of strings into a single string
        """
        finalstring = ''
        for string in stringlist:
            finalstring += string
        return finalstring

    def _dict_from_table(self, captions_columns, captions_lines, array):
        """
        returns a dictionary from a table using specified captions for columns and lines
        """
        return_dict = {}
        for caption_index in range(len(captions_columns)):
            return_dict[captions_columns[caption_index]] = dict(zip(captions_lines, array[:, caption_index]))
        return return_dict

    def _extract_l_valence_charges(self, lines, captions):
        """
        extract the l-decomposed valence charges charges from the output file of a KKRnano run:
        reads the used captions and turns the used tables into a dict
        """
        orbitals = [
            's', 'p', 'd', 'f', 'dummy'
        ]  #KKRnano does not write out "orbitals" beyond this, dummy to simplify accounting for varying lengths
        return_dict = {}

        length_top = search_string(
            '----', lines
        )  #this delimiter marks where only total values follow in the output file, varies w.r.t. used LMAX
        length_bottom = len(lines) - length_top

        #after col 21, only floats follow, that are sought to be read in in the following
        blockstring = self._stringFromList([line[21:] for line in lines[:-length_bottom]])

        used_orbitals = orbitals[:length_top]
        used_orbitals[-1] = 'non-spherical'  #last entry is always the non-spherical part

        array = np.genfromtxt(StringIO(blockstring), dtype=float)
        if array.ndim == 1:
            array = np.transpose(
                np.atleast_2d(array)
            )  #make sure that a 2D array is processed in the following, as this would raise an error otherwise
        # Single String for line where the total values are indicated
        totalstring = self._stringFromList(lines[-length_bottom + 1])

        return_dict = self._dict_from_table(captions, used_orbitals, array)

        totalvalues = np.genfromtxt(StringIO(totalstring[21:]), dtype=float)  #again after col 21, only floats follow

        if np.shape(totalvalues) == ():  #accounting for 0D array
            return_dict['total'] = dict(zip(captions, [totalvalues.item()]))
        else:
            return_dict['total'] = dict(zip(captions, np.genfromtxt(StringIO(totalstring[21:]), dtype=float)))

        return return_dict

    def _convert_all_values_2_python_natives(self, sub_dictionary):
        """
        convert all dtypes of numpy to python native data types in a given (potentially nested) dictionary.
        """
        for key, value in sub_dictionary.items():

            if type(value) is dict:
                self._convert_all_values_2_python_natives(value)

            else:
                try:
                    sub_dictionary[key] = value.item()
                except:
                    pass
        return sub_dictionary

    def _identifyBlocks(self, lines):
        """identify the DOS blocks in the respective output files"""
        blockstartlist = [-1]
        index = 0
        for line in lines:

            block_pos = line.find('&')  #find blocks in DOS file
            if block_pos >= 0:
                blockstartlist.append(index)
            index += 1
        blockstartlist.append(len(lines))  #add EOF line index

        #print(blockstartlist)
        return blockstartlist

    def _get_commentlessLineIndices(self, lines):
        nonemptylines = []  #index list
        index = 0
        for line in lines:
            com_pos = line.find('#')  #comment sign position

            if com_pos < 0:
                nonemptylines.append(index)
            index += 1
        return nonemptylines

    def _process_DOS_file(self, retrieved_folder, filename):
        """reading in DOS file output from KKRnano"""
        data = np.array(self._get_lines(retrieved_folder, filename))
        element = data[0][1:3]

        blocklist = self._identifyBlocks(data)
        dict_dos = {}
        multipleSpins = False
        if len(blocklist) > 1:
            multipleSpins = True
            dict_dos['spin_directions'] = 2
            first_spin = data[0].split(sep='SPIN')[1].split(sep=' ')[1]
            if first_spin == 'DOWN':
                spin_list = ['spin_down', 'spin_up']
            elif first_spin == 'UP':
                spin_list = ['spin_up', 'spin_down']
            else:
                dict_dos['spin_directions'] = 1
        #DOS_blocks=[]

        DOS_captions = ['energy_in_ryd', 's', 'p', 'd', 'f', 'non-spherical', 'total_DOS']
        for i in range(len(blocklist) - 1):
            datablockindices=np.array(self._get_commentlessLineIndices( \
                    data[blocklist[i]+1:blocklist[i+1]-1]))+blocklist[i]+1
            datablockcontent = self._stringFromList(data[datablockindices])
            #print(datablockcontent)
            DOS_block = np.genfromtxt(StringIO(datablockcontent))

            block_dict = {}
            final_DOS_captions = DOS_captions[:np.shape(DOS_block)[1] - 2]
            final_DOS_captions.append(DOS_captions[-2])
            final_DOS_captions.append(DOS_captions[-1])

            for p in range(np.shape(DOS_block)[1]):
                block_dict[final_DOS_captions[p]] = DOS_block[:, p]

            if multipleSpins:
                dict_dos[spin_list[i]] = block_dict  # pylint: disable=used-before-assignment
            else:
                dict_dos = block_dict
            #DOS_blocks.append(DOS_block)
        return element, dict_dos  #DOS_blocks

    def parse(self, debug=False, **kwargs):
        """
        Parse output data folder, store results in database.

        :param retrieved: a dictionary of retrieved nodes, where
          the key is the link name
        :returns: nothing if everything is fine or an exit code defined in the voronoi calculation class
        """

        success = True
        node_list = ()

        # Get retrieved folders
        try:
            retrieved_folder = self.retrieved
            print(retrieved_folder)
        except:
            print('OUT FOLDER NOT FOUND')
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        # check what is inside the folder
        list_of_files = retrieved_folder.list_object_names()

        calc_node = retrieved_folder.get_incoming(node_class=CalcJobNode).first().node

        #Check if a StrucWithPotData object was used as input and take then corresponding structure or
        #if a parent_structure has to be found
        #         if hasattr(calc_node.inputs, 'strucwithpot') and not hasattr(calc_node.inputs, 'parent_folder'):
        #             struc = calc_node.inputs.strucwithpot.structure
        #         else:
        #             struc = find_parent_structure(calc_node)

        #Also for the convert step, these are the files that are supposed to be parsed
        output0_file_name = KKRnanoCalculation._DEFAULT_OUTPUT_PREP_FILE
        output_file_name = KKRnanoCalculation._DEFAULT_OUTPUT_FILE

        # initialize out_dict and parse output files
        out_dict_final = {'parser_version': self._ParserVersion}

        #from --prepare output file
        out0_dict = {}
        #"""
        lines0 = self._get_lines(retrieved_folder, output0_file_name)

        # number of atoms
        try:
            out0_dict['num_atoms'] = int(lines0[search_string('atoms in rbasis.xyz', lines0)].split(sep=' ')[1])

        except:
            print('Number of atoms was not read!')

        # lattice constant

        try:
            alat_string = lines0[search_string('Lattice constants :  ALAT',
                                               lines0)].split(sep=' = ')[1].split(sep='   ')[0].replace(' ', '')
            out0_dict['alat_internal'] = float(alat_string)
            pass
        except:
            print('ALAT was not read!')
        out0_dict['alat_internal_unit'] = 'a_Bohr'

        # Reading k-mesh details

        kmesh_dict = {}
        #kmesh_dict['number_different_kmeshes']=int(lines0[search_string("number of different k-meshes",lines0)].split(sep=" : ")[1])

        try:
            kmesh_dict['number_different_kmeshes'] = int(
                lines0[search_string('number of different k-meshes', lines0)].split(sep=' : ')[1]
            )

        except:
            print('Number of diff. k-meshes was not read!')

        kmesh_caption_KKRnano = 'k-mesh NofKs N kx N ky N kz vol BZ'
        kmesh_caption_aiida_KKRhost = ['number_of_kpts', 'n_kx', 'n_ky', 'n_kz']

        table = self._read_table_block(
            self._find_block(lines0, kmesh_caption_KKRnano, retrieved_folder, output0_file_name), retrieved_folder,
            output0_file_name
        )

        kmesh_dict['number_kpoints_per_kmesh'] = {}

        #filling dictionary with retrieved data
        try:
            for keyindex in range(len(kmesh_caption_aiida_KKRhost)):
                key = kmesh_caption_aiida_KKRhost[keyindex]
                kmesh_dict['number_kpoints_per_kmesh'][key] = np.array(table[:, 1 + keyindex], dtype=int)

            out0_dict['kmesh_group'] = kmesh_dict
        except:
            print('Number of different k-meshes was not read due to unusal format.')

        # reciprocal Bravais matrix
        bravais_caption = 'Reciprocal lattice cell vectors'
        table = self._read_table_block(
            self._find_block([line[:45] for line in lines0], bravais_caption, retrieved_folder, output0_file_name),
            retrieved_folder, output0_file_name
        )

        table = np.array(table[:, 1:], dtype=float)
        out0_dict['reciprocal_bravais_matrix'] = table[:, :3]
        out0_dict['reciprocal_bravais_matrix_unit'] = '2*pi / alat'

        # Bravais matrix
        bravais_caption = 'Direct lattice cell vectors'
        table = self._read_table_block(
            self._find_block([line[:45] for line in lines0], bravais_caption, retrieved_folder, output0_file_name),
            retrieved_folder,
            output0_file_name,
            index_multiple_tables=0
        )

        table = np.array(table[:, 1:], dtype=float)
        out0_dict['direct_bravais_matrix'] = table[:, :3]
        out0_dict['direct_bravais_matrix_unit'] = 'alat'
        #"""

        # read entries from the main output

        stringsInOutputFile = {
            'total_energy_in_ryd': 'TOTAL ENERGY in ryd. :',
            'rms_all_iterations': 'v+ + v-',
            'rms_minus_all_iterations': 'v+ - v-',
            'fermi_energy_in_ryd': 'Fermi energy =',
            'charge_neutrality_in_e': 'charge neutrality in unit cell =',
            'total_magn_moment_in_unit_cell': 'TOTAL mag. moment in unit cell ='
        }

        out_dict = {}
        for key in stringsInOutputFile:
            out_dict[key], _ = self._findSimpleEntries(stringsInOutputFile[key], retrieved_folder, output_file_name)

        out_dict = {
            **out_dict,
            **self._get_total_EnergyeV('total_energy_in_ryd', retrieved_folder, output_file_name, stringsInOutputFile)
        }

        # Get charges in WS cell
        dict_WScell_keys = {
            'charge_in_e': 'charge in Wigner Seitz cell =',
            'spin_moment': 'spin moment in Wigner Seitz cell =',
            'nuclear_charge_in_e': 'nuclear charge',
            'core_charge_in_e': 'core charge'
        }
        WScell_dict = {}
        for key in dict_WScell_keys:
            WScell_dict[key] = self._get_charge_in_WScell(
                dict_WScell_keys[key], retrieved_folder, output_file_name, out0_dict['num_atoms']
            )

        # Extract the l-decomposed valence charges information for all "orbitals"
        # and all iterations and add them to the dict
        lines = self._get_lines(retrieved_folder, output_file_name)

        #find block where the valence charges are indicated
        string2find = 'l-decomposed valence charges'
        indexlist = np.array(self._get_index_list(string2find, retrieved_folder, output_file_name, lines)) + 1

        #using a subdictionary to store the information
        dict_orbitals = {}
        for m in range(len(indexlist)):
            #find iteration block to process
            index = indexlist[m]
            nextindex = indexlist[(m + 1) % len(indexlist)]
            if nextindex <= index:
                nextindex = -1  #use EOF as nextindex, if necessary

            #read in the captions of the table and convert them to a format that is easier to process
            captions_table = np.genfromtxt(StringIO(lines[index + 2]), dtype=str, delimiter='  ')
            captions_table = np.delete(captions_table, np.where(captions_table == ''))
            captions_table = [item.replace(' ', '_') for item in captions_table][1:]

            # removing leading "_" (occurs for some spins)
            for s in range(len(captions_table)):
                if captions_table[s].find('_') == 0:
                    captions_table[s] = captions_table[s][1:]

            #find string list with the lines to process
            blockend = index + 2 + search_string('#########', lines[index + 2:nextindex])
            blocklines = lines[index + 3:blockend]

            #             print("blocklines",blocklines)
            #             print("indexlist", indexlist)
            #             print("index", index)
            #             print("lines", lines[:5])
            #             print("blockend", blockend)
            #             print("nextindex",nextindex)

            atomblocks = np.array(
                self._get_index_list('===', retrieved_folder, lines=blocklines)
            ) + index + 3  #retrieved folder is actually not needed, but is passed for keeping it simple

            if len(atomblocks) > 1:
                atomblocklength = atomblocks[1] - atomblocks[0]
            else:
                atomblocklength = blockend - atomblocks[0]

            #loop over the atomblocks to read in the information for each atom

            dict_atoms = {}
            dict_atoms['atom'] = {}
            for j in range(len(atomblocks)):
                dict_atoms['atom'][j + 1] = self._extract_l_valence_charges(
                    lines[atomblocks[0] + j * atomblocklength + 1:atomblocks[0] + (j + 1) * atomblocklength + 1],
                    captions_table
                )  #all blocks should have the same length

            #add the atom dic-tionary to the one for the iterations
            dict_orbitals[m + 1] = dict_atoms

        # identify DOS-files
        list_of_DOS_files = []
        for file in list_of_files:
            if file.find('DOS') > -1:
                list_of_DOS_files.append(file)

        #process DOS files if necessary
        out_dict_dos = {}
        if len(list_of_DOS_files) > 0:
            for file_index in range(len(list_of_DOS_files)):
                file = list_of_DOS_files[file_index]

                element, DOSblocks = self._process_DOS_file(retrieved_folder, file)
                atomname = f'atom {file_index + 1}'
                out_dict_dos[atomname] = DOSblocks
                out_dict_dos[atomname]['element'] = element

            out_dict_final['DOS'] = out_dict_dos

        dict_orbitals = {'iterations': dict_orbitals}

        out_dict_final['prepare'] = out0_dict
        out_dict_final['WS_charges'] = WScell_dict
        out_dict_final['l_decomposed_charges'] = dict_orbitals

        out_dict_final = {**out_dict, **out_dict_final}

        # create output node and link
        self.out('output_parameters', Dict(dict=out_dict_final))

        if not success:
            return self.exit_codes.ERROR_PARSING_FAILED



# -*- coding: utf-8 -*-
'''
Module with CLI commands for calcjob types of aiida-kkr.
'''
import click
from .launch import launch_voro
from .launch import launch_kkr
#from .launch import launch_kkrimp
from .launch import launch_dos
from .launch import launch_bs
from .launch import launch_scf
#from .launch import launch_gf_writeout
from .launch import launch_kkrimp_scf
#from .launch import launch_kkr_imp_dos
#from .launch import launch_kkr_imp_sub


@click.group('launch')
def cmd_launch():
    """Commands to launch workflows and calcjobs of aiida-kkr."""


# we do it like this and not in short from to avoid cyclic imports
# and get the full bash completion working
cmd_launch.add_command(launch_voro)
cmd_launch.add_command(launch_kkr)
#cmd_launch.add_command(launch_kkrimp)
cmd_launch.add_command(launch_scf)
cmd_launch.add_command(launch_dos)
cmd_launch.add_command(launch_bs)
#cmd_launch.add_command(launch_gf_writeout)
cmd_launch.add_command(launch_kkrimp_scf)
#cmd_launch.add_command(launch_kkr_imp_dos)
#cmd_launch.add_command(launch_kkr_imp_sub)
